# Copyright: (c) 2020, Brian Scholer <@briantist>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import time

from ansible.errors import AnsibleError
from ansible.plugins.action import ActionBase
from ansible.playbook.task import Task
from ansible.utils.display import Display
display = Display()


def clean_async_result(obj):
    for key in ['ansible_job_id', 'ansible_async_watchdog_pid', 'finished', 'results_file', 'started']:
        obj.pop(key, None)
    return obj


class ActionModule(ActionBase):
    _default_async_timeout = 300
    _default_async_poll = 1

    def __init__(self, *args, **kwargs):
        super(ActionModule, self).__init__(*args, **kwargs)

    def run(self, tmp=None, task_vars=None):
        self._supports_check_mode = True
        self._supports_async = False
        check_mode = self._task.check_mode
        async_timeout = self._task.args.get('async_timeout', self._default_async_timeout)
        async_poll = self._task.args.get('async_poll', self._default_async_poll)

        result = super(ActionModule, self).run(tmp, task_vars)

        if async_poll <= 0:
            raise AnsibleError("The 'async_poll' option must be greater than 0, got: %i" % async_poll)

        # build the wait_for_connection object for later use
        wait_for_connection_task = self._task.copy()
        wait_for_connection_task.args = {
            'timeout': async_timeout,
            'sleep': async_poll,
        }
        wait_connection_action = self._shared_loader_obj.action_loader.get(
            'wait_for_connection',
            task=wait_for_connection_task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=self._templar,
            shared_loader_obj=self._shared_loader_obj
        )

        # if it's not in check mode, call the module async so the WinRM restart doesn't kill ansible
        if not check_mode:
            self._task.async_val = async_timeout
            self._task.poll = async_poll

        result = status = self._execute_module(
            task_vars=task_vars,
            module_args=self._task.args
        )
        display.vvvv("Internal Async Result: %r" % status)

        # if we're in check mode (not doing async) return the result now
        if check_mode:
            return result

        # turn off async so we don't run the following actions as async
        self._task.async_val = 0

        # build the async_status object
        async_status_load_params = dict(
            action='async_status jid=%s' % status['ansible_job_id'],
            environment=self._task.environment
        )
        async_status_task = Task().load(async_status_load_params)
        async_status_action = self._shared_loader_obj.action_loader.get(
            'async_status',
            task=async_status_task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=self._templar,
            shared_loader_obj=self._shared_loader_obj
        )

        # build an async_status mode=cleanup object
        async_cleanup_load_params = dict(
            action='async_status mode=cleanup jid=%s' % status['ansible_job_id'],
            environment=self._task.environment
        )
        async_cleanup_task = Task().load(async_cleanup_load_params)
        async_cleanup_action = self._shared_loader_obj.action_loader.get(
            'async_status',
            task=async_cleanup_task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=self._templar,
            shared_loader_obj=self._shared_loader_obj
        )

        # Retries here is a fallback in case the module fails in an unexpected way
        # which can sometimes not properly set the failed field in the return.
        # It is not related to async retries.
        # Without this, that situation would cause an infinite loop.
        max_retries = 3
        retries = 0
        while not check_mode:
            try:
                # check up on the async job
                job_status = async_status_action.run(task_vars=task_vars)
                display.vvvv("Async Job Status: %r" % job_status)

                if job_status.get('failed', False):
                    raise AnsibleError(job_status.get('msg', job_status))

                if job_status.get('finished', False):
                    result = job_status
                    break

                time.sleep(self._task.poll)

            except BaseException as e:
                retries += 1
                if retries >= max_retries:
                    display.vvvv("Max retries reached.")
                    raise e
                display.vvvv("Retrying (%s of %s)" % (retries, max_retries))
                display.vvvv("Falling back to wait_for_connection: %r" % e)
                wait_connection_action.run(task_vars=task_vars)

        try:
            # let's try to clean up after our implicit async
            job_status = async_cleanup_action.run(task_vars=task_vars)
            if job_status.get('failed', False):
                display.vvvv("Clean up of async status failed on the remote host: %r" % job_status.get('msg', job_status))
        except BaseException as e:
            # let's swallow errors during implicit cleanup to aovid interrupting what was otherwise a successful run
            display.vvvv("Clean up of async status failed on the remote host: %r" % e)

        return clean_async_result(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_policies_global_firewall_rulesets_rules
short_description: Resource module for organizations _policies _global _firewall _rulesets
  _rules
description:
  - Manage operations create, update and delete of the resource organizations _policies
    _global _firewall _rulesets _rules.
  - Create an Organization-Wide Policy Firewall Rule.
  - Delete an Organization-Wide Policy Firewall Rule.
  - Update an Organization-Wide Policy Firewall Rule.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  description:
    description: Description of the firewall rule.
    type: str
  destinations:
    description: Destination traffic criteria. Each source or destination bloc is
      capped separately per rule at 100 total segment values. The count is segments_values_count
      the sum of all values across every segment type in that bloc. Ports use a separate
      cap of 100.
    suboptions:
      criteria:
        description: Destination criteria values (not present if 'any' is in matchCriteria).
        suboptions:
          addressRanges:
            description: Address ranges or addresses.
            elements: str
            type: list
          applianceVlans:
            description: Appliance VLANs.
            elements: dict
            suboptions:
              interfaceId:
                description: Interface ID.
                type: str
            type: list
          applicationCategories:
            description: Application categories.
            elements: dict
            suboptions:
              applications:
                description: Applications in this category.
                elements: dict
                suboptions:
                  id:
                    description: Application ID.
                    type: str
                  name:
                    description: Application name.
                    type: str
                type: list
              id:
                description: Category ID.
                type: str
              name:
                description: Category name.
                type: str
            type: list
          applications:
            description: Applications.
            elements: dict
            suboptions:
              id:
                description: Application ID.
                type: str
              name:
                description: Application name.
                type: str
            type: list
          policyObjectGroups:
            description: Policy object groups.
            elements: dict
            suboptions:
              id:
                description: Policy object group ID.
                type: str
            type: list
          policyObjects:
            description: Policy objects.
            elements: dict
            suboptions:
              id:
                description: Policy object ID.
                type: str
            type: list
          ports:
            description: Port numbers or ranges.
            elements: str
            type: list
          services:
            description: Protocol and port services.
            elements: dict
            suboptions:
              ports:
                description: Port numbers or ranges.
                elements: str
                type: list
              protocol:
                description: Protocol (tcp, udp, etc).
                type: str
            type: list
        type: dict
      matchCriteria:
        description: Destination match criteria types.
        elements: str
        type: list
    type: dict
  enabled:
    description: Whether the rule is enabled.
    type: bool
  name:
    description: Name of the firewall rule.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  policy:
    description: Rule policy - allow or deny traffic.
    type: str
  priority:
    description: Rule priority (lower numbers = higher priority).
    type: int
  ruleId:
    description: RuleId path parameter. Rule ID.
    type: str
  rulesetId:
    description: Firewall ruleset ID to associate the rule with.
    type: str
  sources:
    description: Source traffic criteria. Each source or destination bloc is capped
      separately per rule at 100 total segment values. The count is segments_values_count
      the sum of all values across every segment type in that bloc. Ports use a separate
      cap of 100.
    suboptions:
      criteria:
        description: Source criteria values (not present if 'any' is in matchCriteria).
        suboptions:
          addressRanges:
            description: Address ranges or addresses.
            elements: str
            type: list
          applianceVlans:
            description: Appliance VLANs.
            elements: dict
            suboptions:
              interfaceId:
                description: Interface ID.
                type: str
            type: list
          policyObjectGroups:
            description: Policy object groups.
            elements: dict
            suboptions:
              id:
                description: Policy object group ID.
                type: str
            type: list
          policyObjects:
            description: Policy objects.
            elements: dict
            suboptions:
              id:
                description: Policy object ID.
                type: str
            type: list
          ports:
            description: Port numbers or ranges.
            elements: str
            type: list
        type: dict
      matchCriteria:
        description: Source match criteria types.
        elements: str
        type: list
    type: dict
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationPoliciesGlobalFirewallRulesetsRule
    description: Complete reference of the createOrganizationPoliciesGlobalFirewallRulesetsRule
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-policies-global-firewall-rulesets-rule
  - name: Cisco Meraki documentation for organizations deleteOrganizationPoliciesGlobalFirewallRulesetsRule
    description: Complete reference of the deleteOrganizationPoliciesGlobalFirewallRulesetsRule
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-policies-global-firewall-rulesets-rule
  - name: Cisco Meraki documentation for organizations updateOrganizationPoliciesGlobalFirewallRulesetsRule
    description: Complete reference of the updateOrganizationPoliciesGlobalFirewallRulesetsRule
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-policies-global-firewall-rulesets-rule
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_policies_global_firewall_rulesets_rule,
    organizations.Organizations.delete_organization_policies_global_firewall_rulesets_rule,
    organizations.Organizations.update_organization_policies_global_firewall_rulesets_rule,
  - Paths used are
    post /organizations/{organizationId}/policies/global/firewall/rulesets/rules,
    delete /organizations/{organizationId}/policies/global/firewall/rulesets/rules/{ruleId},
    put /organizations/{organizationId}/policies/global/firewall/rulesets/rules/{ruleId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_policies_global_firewall_rulesets_rules:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    description: This is rule 1
    destinations:
      criteria:
        addressRanges:
          - 1.1.1.1
          - 2.2.2.2
        applianceVlans:
          - interfaceId: L_123456789012345678_vlan_200
        applicationCategories:
          - applications:
              - id: meraki:layer7/application/5
                name: Advertising.com
            id: meraki:layer7/category/24
            name: Advertising
        applications:
          - id: meraki:layer7/application/5
            name: Advertising.com
        policyObjectGroups:
          - id: '45'
        policyObjects:
          - id: '23'
        ports:
          - '22'
          - 42-46
        services:
          - ports:
              - '80'
              - '443'
            protocol: tcp
      matchCriteria:
        - addressRanges
        - services
        - applicationCategories
        - applications
        - policyObjects
        - policyObjectGroups
        - applianceVlans
    enabled: true
    name: Allow developers
    organizationId: string
    policy: deny
    priority: 100
    rulesetId: '32'
    sources:
      criteria:
        addressRanges:
          - 1.1.1.1
          - 2.2.2.2
        applianceVlans:
          - interfaceId: L_123456789012345678_vlan_200
        policyObjectGroups:
          - id: '45'
        policyObjects:
          - id: '23'
        ports:
          - '22'
          - 42-46
      matchCriteria:
        - addressRanges
        - ports
        - policyObjects
        - policyObjectGroups
        - applianceVlans
- name: Delete by id
  cisco.meraki.organizations_policies_global_firewall_rulesets_rules:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    organizationId: string
    ruleId: string
- name: Update by id
  cisco.meraki.organizations_policies_global_firewall_rulesets_rules:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    description: This is rule 1
    destinations:
      criteria:
        addressRanges:
          - 1.1.1.1
          - 2.2.2.2
        applianceVlans:
          - interfaceId: L_123456789012345678_vlan_200
        applicationCategories:
          - applications:
              - id: meraki:layer7/application/5
                name: Advertising.com
            id: meraki:layer7/category/24
            name: Advertising
        applications:
          - id: meraki:layer7/application/5
            name: Advertising.com
        policyObjectGroups:
          - id: '45'
        policyObjects:
          - id: '23'
        ports:
          - '22'
          - 42-46
        services:
          - ports:
              - '80'
              - '443'
            protocol: tcp
      matchCriteria:
        - addressRanges
        - services
        - applicationCategories
        - applications
        - policyObjects
        - policyObjectGroups
        - applianceVlans
    enabled: true
    name: Allow developers
    organizationId: string
    policy: deny
    priority: 100
    ruleId: string
    rulesetId: '32'
    sources:
      criteria:
        addressRanges:
          - 1.1.1.1
          - 2.2.2.2
        applianceVlans:
          - interfaceId: L_123456789012345678_vlan_200
        policyObjectGroups:
          - id: '45'
        policyObjects:
          - id: '23'
        ports:
          - '22'
          - 42-46
      matchCriteria:
        - addressRanges
        - ports
        - policyObjects
        - policyObjectGroups
        - applianceVlans
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "ruleId": "string",
      "name": "string",
      "rulesetId": "string",
      "policy": "string",
      "enabled": true,
      "priority": 0,
      "description": "string",
      "sources": {
        "matchCriteria": [
          "string"
        ],
        "criteria": {
          "addressRanges": [
            "string"
          ],
          "ports": [
            "string"
          ],
          "policyObjects": [
            {
              "id": "string"
            }
          ],
          "policyObjectGroups": [
            {
              "id": "string"
            }
          ],
          "applianceVlans": [
            {
              "interfaceId": "string"
            }
          ],
          "siteSpecificVlans": [
            {
              "id": 0,
              "address": {
                "offsets": {
                  "ipv4": 0,
                  "ipv6": "string"
                }
              }
            }
          ]
        }
      },
      "destinations": {
        "matchCriteria": [
          "string"
        ],
        "criteria": {
          "addressRanges": [
            "string"
          ],
          "ports": [
            "string"
          ],
          "services": [
            {
              "protocol": "string",
              "ports": [
                "string"
              ]
            }
          ],
          "applicationCategories": [
            {
              "id": "string"
            }
          ],
          "applications": [
            {
              "id": "string",
              "name": "string"
            }
          ],
          "policyObjects": [
            {
              "id": "string"
            }
          ],
          "policyObjectGroups": [
            {
              "id": "string"
            }
          ],
          "applianceVlans": [
            {
              "interfaceId": "string"
            }
          ],
          "countries": [
            {
              "code": "string"
            }
          ],
          "fqdns": [
            "string"
          ],
          "siteSpecificVlans": [
            {
              "id": 0,
              "address": {
                "offsets": {
                  "ipv4": 0,
                  "ipv6": "string"
                }
              }
            }
          ]
        }
      },
      "createdAt": "string",
      "lastUpdatedAt": "string"
    }
"""



# (c) 2020 CyberArk Software Ltd. All rights reserved.
# (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# pylint: disable=too-many-lines
from __future__ import (absolute_import, division, print_function)

__metaclass__ = type  # pylint: disable=invalid-name

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
    name: conjur_variable
    version_added: "1.0.2"
    short_description: Fetch credentials from CyberArk Conjur.
    author:
      - CyberArk BizDev (@cyberark-bizdev)
    description:
      Retrieves credentials from Conjur using the controlling host's Conjur identity,
      environment variables, or extra-vars.
      Environment variables could be CONJUR_ACCOUNT, CONJUR_APPLIANCE_URL, CONJUR_CERT_FILE,
      CONJUR_CERT_CONTENT, CONJUR_AUTHN_LOGIN, CONJUR_AUTHN_API_KEY, CONJUR_AUTHN_JWT_TOKEN, CONJUR_AUTHN_TOKEN_FILE,
      CONJUR_AUTHN_TYPE, CONJUR_AUTHN_SERVICE_ID, AZURE_CLIENT_ID
      Extra-vars could be conjur_account, conjur_appliance_url, conjur_cert_file, conjur_cert_content,
      conjur_authn_login, conjur_authn_jwt_token, conjur_authn_api_key, conjur_authn_token_file,
      conjur_authn_type, conjur_authn_service_id, azure_client_id
      Conjur info - U(https://www.conjur.org/).
    requirements:
      - 'The controlling host running Ansible has a Conjur identity.
        (More: U(https://docs.conjur.org/latest/en/Content/Get%20Started/key_concepts/machine_identity.html))'
    options:
      _terms:
        description: >
          Variable path to retrieve from Conjur.
          IMPORTANT: Do NOT URL-encode this path. The plugin handles encoding automatically.
          For paths with spaces, use: '/path/to/secret with spaces'
          NOT: '/path/to/secret+with+spaces' or '/path/to/secret%20with%20spaces'
        required: true
      validate_certs:
        description: Flag to control SSL certificate validation
        type: boolean
        default: true
      as_file:
        description: >
          Store lookup result in a temporary file and returns the file path. Thus allowing it to be consumed as an ansible file parameter
          (eg ansible_ssh_private_key_file).
        type: boolean
        default: false
      identity_file:
        description: Path to the Conjur identity file. The identity file follows the netrc file format convention.
        type: path
        default: /etc/conjur.identity
        required: false
        ini:
          - section: conjur,
            key: identity_file_path
        env:
          - name: CONJUR_IDENTITY_FILE
      config_file:
        description: Path to the Conjur configuration file. The configuration file is a YAML file.
        type: path
        default: /etc/conjur.conf
        required: false
        ini:
          - section: conjur,
            key: config_file_path
        env:
          - name: CONJUR_CONFIG_FILE
      conjur_appliance_url:
        description: Conjur appliance url
        type: string
        required: false
        ini:
          - section: conjur,
            key: appliance_url
        vars:
          - name: conjur_appliance_url
        env:
          - name: CONJUR_APPLIANCE_URL
      conjur_authn_login:
        description: Conjur authn login
        type: string
        required: false
        ini:
          - section: conjur,
            key: authn_login
        vars:
          - name: conjur_authn_login
        env:
          - name: CONJUR_AUTHN_LOGIN
      conjur_account:
        description: Conjur account
        type: string
        required: false
        ini:
          - section: conjur,
            key: account
        vars:
          - name: conjur_account
        env:
          - name: CONJUR_ACCOUNT
      conjur_authn_api_key:
        description: Conjur authn api key
        type: string
        required: false
        ini:
          - section: conjur,
            key: authn_api_key
        vars:
          - name: conjur_authn_api_key
        env:
          - name: CONJUR_AUTHN_API_KEY
      conjur_cert_file:
        description: Path to the Conjur cert file
        type: path
        required: false
        ini:
          - section: conjur,
            key: cert_file
        vars:
          - name: conjur_cert_file
        env:
          - name: CONJUR_CERT_FILE
      conjur_cert_content:
        description: Content of the Conjur cert
        type: string
        required: false
        ini:
          - section: conjur,
            key: cert_content
        vars:
          - name: conjur_cert_content
        env:
          - name: CONJUR_CERT_CONTENT
      conjur_authn_token_file:
        description: Path to the access token file
        type: path
        required: false
        ini:
          - section: conjur,
            key: authn_token_file
        vars:
          - name: conjur_authn_token_file
        env:
          - name: CONJUR_AUTHN_TOKEN_FILE
      conjur_authn_type:
        description: Type of Conjur authenticator
        type: string
        required: False
        ini:
          - section: conjur,
            key: authn_type
        vars:
          - name: conjur_authn_type
        env:
          - name: CONJUR_AUTHN_TYPE
      conjur_authn_jwt_token:
        description: Conjur authn-jwt token
        type: string
        required: False
        ini:
          - section: conjur,
            key: authn_jwt_token
        vars:
          - name: conjur_authn_jwt_token
        env:
          - name: CONJUR_AUTHN_JWT_TOKEN
      conjur_authn_jwt_token_file:
        description: >
          Path to a file containing the JWT token used for authn-jwt authentication.
          The file must be readable only by the owner (permissions 0600 or stricter).
          Takes precedence over conjur_authn_jwt_token when both are set.
          The file is not deleted by the plugin after use.
        type: path
        required: False
        ini:
          - section: conjur,
            key: authn_jwt_token_file
        vars:
          - name: conjur_authn_jwt_token_file
        env:
          - name: CONJUR_AUTHN_JWT_TOKEN_FILE
      conjur_authn_service_id:
        description: Service ID for cloud-based authenticators
        type: string
        required: False
        ini:
          - section: conjur,
            key: authn_service_id
        vars:
          - name: conjur_authn_service_id
        env:
          - name: CONJUR_AUTHN_SERVICE_ID
      azure_client_id:
        description: Client id for azure user-assigned managed identity
        type: string
        required: False
        ini:
          - section: azure,
            key: client_id
        vars:
          - name: azure_client_id
        env:
          - name: AZURE_CLIENT_ID
      conjur_client_cert_file:
        description: >
          Path to the client certificate file (PEM format) used for authn-cert
          mutual TLS authentication. Required when conjur_authn_type is set to 'authn-cert'.
          The certificate must be signed by the CA configured for the authn-cert service in Conjur.
        type: path
        required: False
        ini:
          - section: conjur,
            key: client_cert_file
        vars:
          - name: conjur_client_cert_file
        env:
          - name: CONJUR_CLIENT_CERT_FILE
      conjur_client_key_file:
        description: >
          Path to the client private key file (PEM format) used for authn-cert
          mutual TLS authentication. Required when conjur_authn_type is set to 'authn-cert'.
        type: path
        required: False
        ini:
          - section: conjur,
            key: client_key_file
        vars:
          - name: conjur_client_key_file
        env:
          - name: CONJUR_CLIENT_KEY_FILE
      conjur_authn_cert_mode:
        description: >
          Host resolution mode for authn-cert authentication. Valid values are
          'request' (default) and 'spiffe'.
          In 'request' mode the workload identity (conjur_authn_login) must be
          provided and is included in the authentication URL.
          In 'spiffe' mode the workload identity is derived from the X.509 SVID
          SPIFFE URI embedded in the client certificate and conjur_authn_login is
          not required.
        type: str
        required: False
        default: request
        ini:
          - section: conjur,
            key: authn_cert_mode
        vars:
          - name: conjur_authn_cert_mode
        env:
          - name: CONJUR_AUTHN_CERT_MODE
      conjur_authn_jwt_mode:
        description: >
          Identity resolution mode for authn-jwt authentication. Valid values are
          'url' (default) and 'token-app-property'.
          In 'url' mode the workload identity (conjur_authn_login) is included in
          the authentication URL path. CyberArk documentation states this mode
          should not be used in production environments.
          In 'token-app-property' mode the workload identity is derived solely from
          a claim inside the JWT token as configured by the token-app-property
          annotation in the Conjur policy; conjur_authn_login is not required.
        type: str
        required: False
        default: url
        ini:
          - section: conjur,
            key: authn_jwt_mode
        vars:
          - name: conjur_authn_jwt_mode
        env:
          - name: CONJUR_AUTHN_JWT_MODE
      retry_interval:
        description: Time in seconds to wait between retry attempts (default 10)
        type: int
        required: False
        default: 10
        vars:
          - name: conjur_retry_interval
"""

EXAMPLES = """
---
- hosts: localhost
  collections:
    - cyberark.conjur
  tasks:
    - name: Lookup variable in Conjur
      debug:
        msg: "{{ lookup('cyberark.conjur.conjur_variable', '/path/to/secret') }}"
"""

RETURN = """
  _raw:
    description:
      - Value stored in Conjur.
"""

import os
import socket
import traceback
import ssl
import re
import shutil
from base64 import b64encode
from netrc import netrc
from time import sleep
from stat import S_IRUSR, S_IWUSR
from tempfile import gettempdir, NamedTemporaryFile
import atexit
import datetime
import hashlib
import hmac
import json
import urllib.parse
import yaml
import ansible.module_utils.six.moves.urllib.error as urllib_error
from ansible.errors import AnsibleError
from ansible.plugins.lookup import LookupBase
from ansible.module_utils.six.moves.urllib.parse import quote
from ansible.module_utils.urls import open_url
from ansible.utils.display import Display
try:
    from cryptography.x509 import load_pem_x509_certificate
    from cryptography.hazmat.backends import default_backend
except ImportError:
    cryptography_import_error = traceback.format_exc()
else:
    cryptography_import_error = None

display = Display()
temp_cert_file = None
telemetry_header = None


# ************* REQUEST VALUES *************
AWS_METADATA_URL = "http://169.254.169.254/latest/meta-data/iam/security-credentials/"
AWS_AVAILABILITY_ZONE = "http://169.254.169.254/latest/meta-data/placement/availability-zone"
AWS_TOKEN_URL = "http://169.254.169.254/latest/api/token"
METHOD = 'GET'
SERVICE = 'sts'
HOST = 'sts.amazonaws.com'
ENDPOINT = 'https://sts.amazonaws.com'
REQUEST_PARAMETERS = 'Action=GetCallerIdentity&Version=2011-06-15'

AZURE_METADATA_URL = "http://169.254.169.254/metadata/identity/oauth2/token"
GCP_METADATA_URL = "http://metadata/computeMetadata/v1/instance/service-accounts/default/identity"


class ConjurIAMAuthnException(Exception):
    """
    Exception raised when Conjur IAM authentication fails with a 401 Unauthorized error.
    """
    def __init__(self):
        Exception.__init__(
            self,
            "Conjur IAM authentication failed with 401 - Unauthorized. "
            "Check conjur logs for more information"
        )


class IAMRoleNotAvailableException(Exception):
    """
    Raised when the IAM Role is not available or incorrectly configured.
    """
    def __init__(self):
        Exception.__init__(
            self,
            "Most likely the ec2 instance is configured with no or an incorrect iam role"
        )


class InvalidAwsAccountIdException(Exception):
    """
    Raised when the AWS Account ID is invalid
    """
    def __init__(self):
        Exception.__init__(
            self,
            "The AWS Account ID specified in the CONJUR_AUTHN_LOGIN is invalid and "
            "must be a 12 digit number"
        )


def _valid_aws_account_number(host_id):
    """
    Checks if the given host_id contains a valid 12-digit AWS Account ID.
    """
    parts = host_id.split("/")
    account_id = parts[len(parts) - 2]
    if len(account_id) == 12:
        return True
    return False


def _sign(key, msg):
    """
    Signs the given message using the provided key and HMAC with SHA-256.
    """
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()


def _get_signature_key(key, date_stamp, region_name, service_name):
    """
    Generates the AWS V4 signing key using the provided key,
    date, region, and service name.
    """
    k_date = _sign(('AWS4' + key).encode('utf-8'), date_stamp)
    k_region = _sign(k_date, region_name)
    k_service = _sign(k_region, service_name)
    k_signing = _sign(k_service, 'aws4_request')
    return k_signing


def _get_aws_region():
    """
    Get the AWS Region.
    """
    return "us-east-1"


def _get_iam_role_name():
    """
    Retrieves the IAM Role Name associated with the current environment.
    """
    token = _get_metadata_token()
    headers = {}
    if token:
        headers = {'X-aws-ec2-metadata-token': token}
    res = open_url(
        AWS_METADATA_URL,
        method='GET',
        validate_certs=False,
        ca_path=None,
        headers=headers
    )
    res_body = res.read().decode('utf-8')
    return res_body


def _get_metadata_token():
    """Request a session token for IMDSv2"""
    headers = {'X-aws-ec2-metadata-token-ttl-seconds': '900'}
    token = None
    try:
        response = open_url(
            AWS_TOKEN_URL,
            method='PUT',
            validate_certs=False,
            ca_path=None,
            headers=headers
        )
        response_body = response.read().decode('utf-8')

        if response.getcode() == 200:
            return response_body
        return None
    except Exception as error:  # pylint: disable=broad-except
        display.warning(f"IMDSv2 token retrieval failed: {str(error)}. Falling back to IMDSv1.")
        return None
    finally:
        if token:
            token = None


def _get_iam_role_metadata(role_name, token=None):
    """
    Retrieves metadata for the IAM role associated with the current environment.
    """

    headers = {}
    if token:
        headers = {'X-aws-ec2-metadata-token': token}

    try:
        res = open_url(
            AWS_METADATA_URL + role_name,
            method='GET',
            headers=headers,
            validate_certs=False
        )
        if res.getcode() == 404:
            raise AnsibleError(f"Error retrieving IAM role metadata: {str(res.getcode())}")

        if res.getcode() != 200:
            raise AnsibleError(f"Error retrieving IAM role metadata: {str(res.getcode())}")

        res_body = res.read().decode('utf-8')
        json_dict = json.loads(res_body)

        access_key_id = json_dict["AccessKeyId"]
        secret_access_key = json_dict["SecretAccessKey"]
        token = json_dict["Token"]

        return access_key_id, secret_access_key, token

    except Exception as error:
        raise AnsibleError(f"Error retrieving IAM role metadata: Exception occurred - {str(error)}") from error


def _create_canonical_request(amzdate, token, signed_headers, payload_hash):
    """
    Creates the canonical request string for signing the AWS request.
    """
    canonical_uri = '/'
    canonical_querystring = REQUEST_PARAMETERS
    canonical_headers = 'host:' + HOST + '\n' + 'x-amz-content-sha256:' + payload_hash + '\n' + \
        'x-amz-date:' + amzdate + '\n' + 'x-amz-security-token:' + token + '\n'

    canonical_request = METHOD + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + \
        canonical_headers + '\n' + signed_headers + '\n' + payload_hash

    return canonical_request


# pylint: disable=too-many-arguments,too-many-locals
def _create_conjur_iam_api_key(iam_role_name=None, access_key=None, secret_key=None, token=None):
    """
    Creates an IAM API key for Conjur authentication using the provided IAM role and credentials.
    """
    if iam_role_name is None:
        iam_role_name = _get_iam_role_name()

    metadata_token = _get_metadata_token()

    if access_key is None and secret_key is None and token is None:
        access_key, secret_key, token = _get_iam_role_metadata(iam_role_name, metadata_token)

    region = _get_aws_region()

    if access_key is None or secret_key is None:
        raise AnsibleError('No access key is available.')

    date = datetime.datetime.now(datetime.timezone.utc)
    amzdate = date.strftime('%Y%m%dT%H%M%SZ')
    datestamp = date.strftime('%Y%m%d')

    signed_headers = 'host;x-amz-content-sha256;x-amz-date;x-amz-security-token'
    payload_hash = hashlib.sha256(('').encode('utf-8')).hexdigest()
    canonical_request = _create_canonical_request(amzdate, token, signed_headers, payload_hash)

    algorithm = 'AWS4-HMAC-SHA256'
    credential_scope = datestamp + '/' + region + '/' + SERVICE + '/' + 'aws4_request'
    string_to_sign = algorithm + '\n' + amzdate + '\n' + credential_scope + \
        '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()

    signing_key = _get_signature_key(secret_key, datestamp, region, SERVICE)

    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

    authorization_header = (
        algorithm + ' ' +
        'Credential=' + access_key + '/' + credential_scope + ', ' +
        'SignedHeaders=' + signed_headers + ', ' +
        'Signature=' + signature
    )

    headers = {
        'host': HOST,
        'x-amz-date': amzdate,
        'x-amz-security-token': token,
        'x-amz-content-sha256': payload_hash,
        'authorization': authorization_header
    }

    access_key = None
    secret_key = None
    token = None
    signing_key = None
    string_to_sign = None
    canonical_request = None

    return f'{headers}'.replace("'", '"')


def _fetch_conjur_iam_session_token(
    appliance_url, account, service_id, host_id, cert_file, validate_certs,
    iam_role_name=None, access_key=None, secret_key=None, token=None
):
    """
    Retrieves the Conjur IAM session token for the provided service and IAM role credentials.
    """

    if not _valid_aws_account_number(host_id):
        raise InvalidAwsAccountIdException()

    # Get the telemetry header
    encoded_telemetry = _telemetry_header()

    headers = {
        'x-cybr-telemetry': encoded_telemetry
    }

    appliance_url = appliance_url.rstrip("/")
    url = (
        f"{appliance_url}/authn-iam/{service_id}/{account}/"
        f"{urllib.parse.quote(host_id, safe='')}/authenticate"
    )

    iam_api_key = _create_conjur_iam_api_key(iam_role_name, access_key, secret_key, token)

    try:
        res = open_url(
            url,
            data=iam_api_key,
            method='POST',
            validate_certs=validate_certs,
            ca_path=cert_file,
            headers=headers
        )
        res_body = res.read()
        if res.getcode() == 401:
            raise ConjurIAMAuthnException()
        if res.getcode() != 200:
            raise AnsibleError(f'Failed to authenticate - (got {str(res.getcode())} response)')

        return res_body
    finally:
        iam_api_key = None


def _validate_pem_certificate(cert_content):
    # Normalize line endings
    if '\r\n' in cert_content:
        cert_content = cert_content.replace('\r\n', '\n').strip()
    elif '\r' in cert_content:
        cert_content = cert_content.replace('\r', '\n').strip()
    cert_content = re.sub(r'^[ \t]+', '', cert_content, flags=re.M)
    cert_content = re.sub(r'[ \t]+$', '', cert_content, flags=re.M)
    cert_content = re.sub(r'\n+', '\n', cert_content)

    if not re.match(r"^-----BEGIN CERTIFICATE-----.+-----END CERTIFICATE-----$", cert_content, re.DOTALL):
        raise AnsibleError("Invalid Certificate format.")

    try:
        load_pem_x509_certificate(cert_content.encode(), default_backend())
        return cert_content
    except ValueError as err:
        raise AnsibleError(
            f"Invalid certificate content provided: {str(err)}. "
            "Please check the certificate format."
        ) from err
    except ssl.SSLError as err:
        raise AnsibleError(
            f"SSL error while validating the certificate: {str(err)}. "
            "The certificate may be corrupted or invalid."
        ) from err
    except Exception as err:
        raise AnsibleError(
            f"An error occurred while validating the certificate: {str(err)}. "
            "Please verify the certificate format and try again."
        ) from err


def _get_valid_certificate(cert_content, cert_file):
    if cert_content:
        try:
            display.vvv("Validating provided certificate content")
            cert_content = _validate_pem_certificate(cert_content)
            return cert_content
        except AnsibleError as err:
            display.warning(f"Invalid certificate content: {str(err)}. Attempting to use certificate file.")

    # If cert_content is invalid or missing, fall back to cert_file
    if cert_file:
        if not os.path.exists(cert_file):
            raise AnsibleError(f"Certificate file `{cert_file}` does not exist or cannot be found.")
        try:
            with open(cert_file, 'rb') as file:
                cert_file_content = file.read().decode('utf-8')
                cert_file_content = _validate_pem_certificate(cert_file_content)
                return cert_file_content
        except Exception as err:
            raise AnsibleError(f"Failed to load or validate certificate file `{cert_file}`: {str(err)}") from err

    # If both cert_content and cert_file are missing or invalid, raise an error
    raise AnsibleError("Both certificate content and certificate file are invalid or missing. Please provide a valid certificate.")


def _get_certificate_file(cert_content, cert_file):
    """
    Creates a CA bundle containing CAs from the system truststore appended with the
    provided certificate. The CA bundle is stored in a temporary file.

    The file must persist until the end of the process, so `delete=False` is used
    to prevent automatic deletion. We manually handle deletion at the end of the process
    to ensure proper resource management. A global variable tracks the file for cleanup.

    Pylint's warning about not using a `with` statement is disabled because the file
    needs to remain accessible beyond the function scope.

    Args:
        cert_content (str): Raw certificate content.
        cert_file (str): Path to an existing certificate file, if provided.

    Returns:
        str: Path to the certificate file to be used.
    """
    global temp_cert_file
    cert_content = _get_valid_certificate(cert_content, cert_file)

    if cert_content:
        try:
            temp_cert_file = NamedTemporaryFile(delete=False)  # pylint: disable=consider-using-with

            # Prepare a CA bundle including system CA certificates
            system_ca_bundle = ssl.get_default_verify_paths().cafile
            if system_ca_bundle and os.path.exists(system_ca_bundle):
                with open(system_ca_bundle, 'rb') as bundle:
                    shutil.copyfileobj(bundle, temp_cert_file)
            else:
                display.warning("System CA bundle not found, only the provided cert(s) will be used.")

            # Normalize and append custom cert(s)
            cert_bytes = cert_content.encode() if isinstance(cert_content, str) else cert_content
            temp_cert_file.write(b"\n" + cert_bytes.strip().replace(b"\r\n", b"\n") + b"\n")
            temp_cert_file.close()
            cert_file = temp_cert_file.name
        except Exception as err:
            raise AnsibleError(f"Failed to create temporary CA bundle: {str(err)}") from err

    return cert_file


# Load configuration and return as dictionary if file is present on file system
def _load_conf_from_file(conf_path):
    display.vvv(f'conf file: {conf_path}')

    if not os.path.exists(conf_path):
        return {}
        # raise AnsibleError('Conjur configuration file `{conf_path}` was not found on the controlling host')

    display.vvvv(f'Loading configuration from: {conf_path}')
    with open(conf_path, encoding="utf-8") as file:
        config = yaml.safe_load(file.read())
        return config


# Load identity and return as dictionary if file is present on file system
def _load_identity_from_file(identity_path, appliance_url):
    display.vvvv(f'identity file: {identity_path}')

    if not os.path.exists(identity_path):
        return {}
        # raise AnsibleError(f'Conjur identity file `{identity_path}` was not found on the controlling host')

    display.vvvv(f'Loading identity from: {identity_path} for {appliance_url}')

    conjur_authn_url = f'{appliance_url}/authn'
    identity = netrc(identity_path)

    if identity.authenticators(conjur_authn_url) is None:
        raise AnsibleError(f'The netrc file on the controlling host does not contain an entry for: {conjur_authn_url}')

    host_id, unused, api_key = identity.authenticators(conjur_authn_url)  # pylint: disable=unused-variable

    if not host_id or not api_key:
        return {}

    return {'id': host_id, 'api_key': api_key}


# Merge multiple dictionaries by using dict.update mechanism
def _merge_dictionaries(*arg):
    ret = {}
    for item in arg:
        ret.update(item)
    return ret


# The `quote` method's default value for `safe` is '/' so it doesn't encode slashes
# into "%2F" which is what the Conjur server expects. Thus, we need to use this
# method with no safe characters. We can't use the method `quote_plus` (which encodes
# slashes correctly) because it encodes spaces into the character '+' instead of "%20"
# as expected by the Conjur server
def _encode_str(input_str):
    return quote(input_str, safe='')


# Prepare the telemetry header
def _telemetry_header():
    global telemetry_header

    if telemetry_header is None:
        plugin_dir = os.path.dirname(__file__)
        collection_root = os.path.abspath(os.path.join(plugin_dir, '..', '..'))
        version_file_path = os.path.join(collection_root, 'VERSION')
        if os.path.isfile(version_file_path):
            with open(version_file_path, 'r', encoding='utf-8') as version_file:
                version = version_file.read().strip()
        else:
            version_file_path = os.path.join(collection_root, 'collections', 'ansible_collections', 'cyberark', 'conjur', 'VERSION')
            if os.path.isfile(version_file_path):
                with open(version_file_path, 'r', encoding='utf-8') as version_file:
                    version = version_file.read().strip()
            else:
                galaxy_file_path = os.path.join(collection_root, 'galaxy.yml')
                if os.path.isfile(galaxy_file_path):
                    try:
                        with open(galaxy_file_path, 'r', encoding='utf-8') as galaxy_file:
                            galaxy_data = yaml.safe_load(galaxy_file)
                            version = galaxy_data['version']
                    except (IOError, KeyError, yaml.YAMLError):
                        version = "unknown"
                else:
                    version = "unknown"

        # Prepare the telemetry value
        telemetry_val = f'in=Ansible Collections&it=cybr-secretsmanager&iv={version}&vn=Ansible'

        # Encode to base64
        telemetry_header = b64encode(telemetry_val.encode()).decode().rstrip("=")
    return telemetry_header


# Use credentials to retrieve temporary authorization token
def _fetch_conjur_token(conjur_url, account, username, api_key, validate_certs, cert_file):  # pylint: disable=too-many-arguments
    conjur_url = f'{conjur_url}/authn/{account}/{_encode_str(username)}/authenticate'
    display.vvvv(f'Authentication request to Conjur at: {conjur_url}, with user: {_encode_str(username)}')

    # Get the telemetry header
    encoded_telemetry = _telemetry_header()

    # Prepare headers
    headers = {
        'x-cybr-telemetry': encoded_telemetry
    }

    response = open_url(conjur_url,
                        data=api_key,
                        method='POST',
                        validate_certs=validate_certs,
                        ca_path=cert_file,
                        headers=headers)
    code = response.getcode()
    if code != 200:
        raise AnsibleError(f'Failed to authenticate as \'{username}\' (got {code} response)')

    return response.read()


def _read_jwt_token_file(token_file):
    """Read a JWT token from a file, validating that the file has owner-only permissions."""
    if not os.path.exists(token_file):
        raise AnsibleError(
            f"JWT token file '{token_file}' does not exist. "
            "It is required when conjur_authn_jwt_token_file is set."
        )
    file_mode = os.stat(token_file).st_mode & 0o777
    if file_mode & 0o177:
        display.warning(
            f"JWT token file '{token_file}' has insecure permissions ({oct(file_mode)}). "
            "The file should be readable only by the owner (chmod 600)."
        )
    try:
        with open(token_file, 'r', encoding='utf-8') as fh:
            return fh.read().strip()
    except OSError as err:
        raise AnsibleError(
            f"Failed to read JWT token file '{token_file}': {err}"
        ) from err


def _fetch_conjur_jwt_token(
    appliance_url, account, service_id,
    host_id, jwt_token, cert_file, validate_certs
):
    # Get the telemetry header
    encoded_telemetry = _telemetry_header()

    # Prepare headers
    headers = {
        'x-cybr-telemetry': encoded_telemetry
    }

    try:
        appliance_url = appliance_url.rstrip("/")
        encoded_service_id = urllib.parse.quote(service_id, safe='')
        encoded_account = urllib.parse.quote(account, safe='')
        # token-app-property mode: host_id is None — identity comes from JWT claim mapping
        # url mode: host_id is included in the path
        if host_id:
            url = (
                f"{appliance_url}/authn-jwt/{encoded_service_id}/"
                f"{encoded_account}/"
                f"{urllib.parse.quote(host_id, safe='')}/authenticate"
            )
        else:
            url = f"{appliance_url}/authn-jwt/{encoded_service_id}/{encoded_account}/authenticate"

        token = f"jwt={jwt_token}"

        response = open_url(
            url,
            method='POST',
            data=token,
            headers=headers,
            validate_certs=validate_certs,
            ca_path=cert_file,
            timeout=10
        )
        if response.getcode() != 200:
            raise AnsibleError(f"Error authenticating with Conjur: HTTP {str(response.getcode())}")
        return response.read()

    except urllib_error.URLError as error:
        raise AnsibleError(f"Error fetching identity token: URL error occurred - {str(error)}") from error

    except RuntimeError as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error

    except Exception as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error
    finally:
        token = None
        jwt_token = None


def retry(retries, retry_interval):
    """
    Custom retry decorator

    Args:
        retries (int, optional): Number of retries. Defaults to 5.
        retry_interval (int, optional): Time to wait between intervals. Defaults to 10.
    """
    def parameters_wrapper(target):
        def decorator(*args, **kwargs):
            retry_count = 0
            while True:
                retry_count += 1
                try:
                    return target(*args, **kwargs)
                except (socket.timeout, urllib_error.URLError) as e:
                    # Only retry network/connectivity errors - other errors should fail-fast
                    display.v(
                        f'Retryable error encountered: {type(e).__name__} - {e}. '
                        f'Retrying ({retry_count}/{retries})...'
                    )
                    if retry_count >= retries:
                        raise
                    sleep(retry_interval)
        return decorator
    return parameters_wrapper


def _repeat_open_url(url, headers=None, method=None, validate_certs=True, ca_path=None, retry_interval=10):
    """
    Wrapper for open_url with retry logic

    Args:
        retry_interval: Time in seconds between retry attempts (default: 10)
    """
    @retry(retries=5, retry_interval=retry_interval)
    def _do_request():
        return open_url(url,
                        headers=headers,
                        method=method,
                        validate_certs=validate_certs,
                        ca_path=ca_path)
    return _do_request()


# Retrieve Conjur variable using the temporary token
def _fetch_conjur_variable(conjur_variable, token, conjur_url, account, validate_certs, cert_file, retry_interval=10):  # pylint: disable=too-many-arguments
    token = b64encode(token)
    # Get the telemetry header
    encoded_telemetry = _telemetry_header()

    headers = {
        'Authorization': f'Token token="{token.decode("utf-8")}"',
        'x-cybr-telemetry': encoded_telemetry
    }

    url = f'{conjur_url}/secrets/{account}/variable/{_encode_str(conjur_variable)}'
    display.vvvv(f'Conjur Variable URL: {url}')

    response = _repeat_open_url(url,
                                headers=headers,
                                method='GET',
                                validate_certs=validate_certs,
                                ca_path=cert_file,
                                retry_interval=retry_interval)

    if response.getcode() == 200:
        display.vvvv(f'Conjur variable {conjur_variable} was successfully retrieved')
        value = response.read().decode("utf-8")
        return [value]
    if response.getcode() == 401:
        raise AnsibleError('Conjur request has invalid authorization credentials')
    if response.getcode() == 403:
        raise AnsibleError(f'The controlling host\'s Conjur identity does not have authorization to retrieve {conjur_variable}')
    if response.getcode() == 404:
        raise AnsibleError(f'The variable {conjur_variable} does not exist')

    return {}


def _default_tmp_path():
    if os.access("/dev/shm", os.W_OK):
        return "/dev/shm"

    return gettempdir()


def _store_secret_in_file(value):
    """
    Writes a secret value to a secure temporary file and returns its path.

    The file is created in /dev/shm or /tmp (based on `_default_tmp_path()`),
    with user-only read/write permissions. `delete=False` ensures the file
    persists beyond this function, as it needs to be accessible later.

    Note: We avoid using a `with` statement here to prevent premature file
    closure or deletion, which would make the file unusable.

    Args:
        value (list): List containing the secret string.

    Returns:
        list: Path to the temporary file as a single-item list.
    """
    secrets_file = NamedTemporaryFile(mode='w', dir=_default_tmp_path(), delete=False)  # pylint: disable=consider-using-with
    atexit.register(lambda p=secrets_file.name: os.path.exists(p) and os.unlink(p))
    os.chmod(secrets_file.name, S_IRUSR | S_IWUSR)
    secrets_file.write(value[0])
    return [secrets_file.name]


# Fetch token using Conjur Certificate Authenticator (authn-cert)
def _validate_authn_cert_files(client_cert_file, client_key_file):
    """Raise AnsibleError if the client certificate or key file is missing."""
    if not client_cert_file or not os.path.exists(client_cert_file):
        raise AnsibleError(
            f"Client certificate file '{client_cert_file}' is missing or does not exist. "
            "It is required for authn-cert authentication."
        )
    if not client_key_file or not os.path.exists(client_key_file):
        raise AnsibleError(
            f"Client key file '{client_key_file}' is missing or does not exist. "
            "It is required for authn-cert authentication."
        )
    key_mode = os.stat(client_key_file).st_mode & 0o777
    if key_mode & 0o177:
        display.warning(
            f"Client key file '{client_key_file}' has insecure permissions ({oct(key_mode)}). "
            "The file should be readable only by the owner (chmod 600)."
        )


def _fetch_conjur_cert_token(
    appliance_url, account, service_id, host_id,
    client_cert_file, client_key_file,
    ca_cert_file, validate_certs
):
    """
    Authenticates to Conjur using the authn-cert certificate authenticator.

    API reference:
      POST {appliance_url}/authn-cert/{service_id}/{account}[/{workload_id}]/authenticate

    The client certificate is passed as the ``X-SSL-Client-Certificate`` HTTP
    header (percent-encoded), matching Conjur's expected format when deployed
    behind a TLS-terminating proxy (e.g. nginx).  This mirrors the Ruby
    ``CGI.escape(client_cert.to_pem)`` pattern used by the Conjur server tests.

    Two host-resolution modes are supported (configured server-side):
      - request (default): workload identity is taken from the URL path
        (``host_id`` must be provided, e.g. ``host/my-policy/my-host``).
      - spiffe: workload identity is derived from the SPIFFE URI in the
        client certificate's SAN field; ``host_id`` must be ``None``.

    Args:
        appliance_url (str):     Base URL of the Conjur appliance.
        account (str):           Conjur account name.
        service_id (str):        Service ID of the authn-cert authenticator.
        host_id (str|None):      Full Conjur host identity for request mode,
                                 or None for SPIFFE mode.
        client_cert_file (str):  Path to the client PEM certificate file.
        client_key_file (str):   Path to the client PEM private key file
                                 (currently unused for header-based auth but
                                 validated for presence).
        ca_cert_file (str|None): Path to Conjur server CA/bundle cert file.
        validate_certs (bool):   Whether to validate the server TLS certificate.

    Returns:
        bytes: The raw Conjur access token bytes.

    Raises:
        AnsibleError: On missing files, authentication failure, or HTTP errors.
    """
    _validate_authn_cert_files(client_cert_file, client_key_file)

    try:
        with open(client_cert_file, 'r', encoding='utf-8') as cert_fh:
            client_cert_pem = cert_fh.read()
    except OSError as err:
        raise AnsibleError(
            f"Failed to read client certificate file '{client_cert_file}': {err}"
        ) from err

    encoded_cert = urllib.parse.quote(client_cert_pem, safe='')
    appliance_url = appliance_url.rstrip("/")
    encoded_service_id = urllib.parse.quote(service_id, safe='')
    encoded_account = urllib.parse.quote(account, safe='')

    # host_id is required for 'request' mode (default) and absent for 'spiffe' mode.
    # Per API docs: POST {url}/authn-cert/{service_id}/{account}[/{workload_id}]/authenticate
    if host_id:
        url = (
            f"{appliance_url}/authn-cert/{encoded_service_id}/{encoded_account}/"
            f"{urllib.parse.quote(host_id, safe='')}/authenticate"
        )
    else:
        url = f"{appliance_url}/authn-cert/{encoded_service_id}/{encoded_account}/authenticate"

    headers = {
        'X-SSL-Client-Certificate': encoded_cert,
        'x-cybr-telemetry': _telemetry_header()
    }
    display.vvv(
        f'authn-cert: POST {url} '
        f'[cert={client_cert_file}, key={client_key_file}, ca={ca_cert_file}]'
    )

    try:
        res = open_url(
            url,
            data=b'',
            method='POST',
            headers=headers,
            validate_certs=validate_certs,
            ca_path=ca_cert_file,
            client_cert=client_cert_file,
            client_key=client_key_file,
        )
        if res.getcode() != 200:
            raise AnsibleError(
                f'Failed to authenticate with certificate (got {str(res.getcode())} response)'
            )
        return res.read()
    except AnsibleError:
        raise
    except urllib_error.URLError as error:
        raise AnsibleError(
            f"Error during authn-cert authentication: {error}"
        ) from error
    except ssl.SSLError as error:
        raise AnsibleError(
            f"Error during authn-cert authentication: SSL error — {error}"
        ) from error
    except OSError as error:
        raise AnsibleError(
            f"Error during authn-cert authentication: {error}"
        ) from error
    except Exception as error:  # pylint: disable=broad-except
        raise AnsibleError(
            f"Error during authn-cert authentication: {error}"
        ) from error


# Fetch token from aure vm, func, app and authn with conjur for access token
def _fetch_conjur_azure_token(
    appliance_url, account, service_id,
    host_id, cert_file, validate_certs, client_id=""
):
    try:
        params = {
            "api-version": "2018-02-01",
            "resource": "https://management.azure.com/"
        }

        if client_id:
            params["client_id"] = client_id

        headers = {
            "Metadata": "true"
        }
        url_with_params = f"{AZURE_METADATA_URL}?{urllib.parse.urlencode(params)}"
        response = open_url(
            url_with_params,
            method='GET',
            headers=headers,
            validate_certs=validate_certs,
            timeout=10
        )

        response_body = response.read().decode('utf-8')
        # Parse JSON response
        data = json.loads(response_body)

        if response.getcode() != 200:
            raise AnsibleError(f"Error retrieving token from azure: {str(response.getcode())}")

        appliance_url = appliance_url.rstrip("/")
        url = (
            f"{appliance_url}/authn-azure/{service_id}/{account}/"
            f"{urllib.parse.quote(host_id, safe='')}/authenticate"
        )

        token = f"jwt={data.get('access_token')}"

        # Get the telemetry header
        encoded_telemetry = _telemetry_header()

        # Prepare headers
        headers = {
            'x-cybr-telemetry': encoded_telemetry
        }
        response = open_url(
            url,
            method='POST',
            data=token.encode('utf-8'),
            headers=headers,
            validate_certs=validate_certs,
            ca_path=cert_file,
            timeout=10
        )
        if response.getcode() != 200:
            raise AnsibleError(f"Error authenticating with Conjur: HTTP {str(response.getcode())}")
        return response.read()

    except urllib.error.URLError as error:
        raise AnsibleError(f"Error fetching identity token: URL error occurred - {str(error)}") from error

    except RuntimeError as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error

    except Exception as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error
    finally:
        client_id = None
        token = None
        response_body = None


def _fetch_conjur_gcp_identity_token(
    appliance_url, account, host_id, cert_file, validate_certs
):
    try:
        params = {
            'audience': f'conjur/{account}/{host_id}',
            'format': 'full'
        }
        headers = {'Metadata-Flavor': 'Google'}

        url_with_params = f"{GCP_METADATA_URL}?{urllib.parse.urlencode(params)}"

        response = open_url(
            url_with_params,
            method='GET',
            headers=headers,
            validate_certs=validate_certs,
            timeout=10
        )

        if response.getcode() != 200:
            raise AnsibleError(f"Error retrieving token from gcp: {str(response.getcode())}")

        response_body = response.read().decode('utf-8')
        appliance_url = appliance_url.rstrip("/")
        url = f"{appliance_url}/authn-gcp/{account}/authenticate"

        token = f"jwt={response_body}"

        # Get the telemetry header
        encoded_telemetry = _telemetry_header()

        # Prepare headers
        headers = {
            'x-cybr-telemetry': encoded_telemetry
        }
        response = open_url(
            url,
            method='POST',
            data=token.encode('utf-8'),
            headers=headers,
            validate_certs=validate_certs,
            ca_path=cert_file,
            timeout=10
        )
        if response.getcode() != 200:
            raise AnsibleError(f"Error: Received status code {str(response.getcode())}")

        return response.read()
    except urllib.error.URLError as error:
        raise AnsibleError(f"Error fetching identity token: URL error occurred - {str(error)}") from error

    except RuntimeError as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error

    except Exception as error:
        raise AnsibleError(f"Error fetching identity token: {str(error)}") from error
    finally:
        token = None
        response_body = None


class LookupModule(LookupBase):

    def run(self, terms, variables=None, **kwargs):  # pylint: disable=too-many-locals,missing-function-docstring,too-many-branches,too-many-statements
        if terms == []:
            raise AnsibleError("Invalid secret path: no secret path provided.")
        if not terms[0] or terms[0].isspace():
            raise AnsibleError("Invalid secret path: empty secret path not accepted.")

        # We should register the variables as LookupModule options.
        #
        # Doing this has some nice advantages if we're considering supporting
        # a set of Ansible variables that could sometimes replace environment
        # variables.
        #
        # Registering the variables as options forces them to adhere to the
        # behavior described in the DOCUMENTATION variable. An option can have
        # both a Ansible variable and environment variable source, which means
        # Ansible will do some juggling on our behalf.
        self.set_options(var_options=variables, direct=kwargs)

        appliance_url = self.get_var_value("conjur_appliance_url")
        account = self.get_var_value("conjur_account")
        authn_login = self.get_var_value("conjur_authn_login")
        authn_api_key = self.get_var_value("conjur_authn_api_key")
        cert_file = self.get_var_value("conjur_cert_file")
        cert_content = self.get_var_value("conjur_cert_content")
        authn_token_file = self.get_var_value("conjur_authn_token_file")
        authn_type = self.get_var_value("conjur_authn_type")
        service_id = self.get_var_value("conjur_authn_service_id")
        azure_client_id = self.get_var_value("azure_client_id")
        jwt_token = self.get_var_value("conjur_authn_jwt_token")
        jwt_token_file = self.get_var_value("conjur_authn_jwt_token_file")
        client_cert_file = self.get_var_value("conjur_client_cert_file")
        client_key_file = self.get_var_value("conjur_client_key_file")
        cert_mode = self.get_var_value("conjur_authn_cert_mode") or "request"
        jwt_mode = self.get_var_value("conjur_authn_jwt_mode") or "url"
        retry_interval = self.get_option('retry_interval')

        validate_certs = self.get_option('validate_certs')
        conf_file = self.get_option('config_file')
        as_file = self.get_option('as_file')

        if validate_certs is False:
            display.warning('Certificate validation has been disabled. Please enable with validate_certs option.')

        if 'http://' in str(appliance_url):
            raise AnsibleError(('[WARNING]: Conjur URL uses insecure connection. Please consider using HTTPS.'))

        if validate_certs is True:
            cert_file = _get_certificate_file(cert_content, cert_file)

        if authn_type in ("aws", "azure", "jwt", "authn-cert") and service_id is None:
            raise AnsibleError("[WARNING]: Please set the conjur_authn_service_id for AWS, Azure, JWT or authn-cert authenticator")

        if not account:
            display.vvv("No conjur account provided. Defaulting to 'conjur'.")
            account = "conjur"

        conf = _merge_dictionaries(
            _load_conf_from_file(conf_file),
            {
                "account": account,
                "appliance_url": appliance_url
            } if (appliance_url is not None)
            else {},
            {
                "cert_file": cert_file
            } if (cert_file is not None)
            else {},
            {
                "authn_token_file": authn_token_file
            } if authn_token_file is not None
            else {}
        )

        if 'appliance_url' not in conf:
            raise AnsibleError(
                """Configuration must define options `conjur_appliance_url`.
                This config can be set by any of the following methods, listed in order of priority:
                - Ansible variable `conjur_appliance_url`, set in the parent playbook or
                  passed via --extra-vars
                - Environment variable `CONJUR_APPLIANCE_URL`
                - A configuration file on the controlling host with the field `appliance_url`"""
            )

        if 'authn_token_file' not in conf:
            identity_file = self.get_option('identity_file')
            identity = _merge_dictionaries(
                _load_identity_from_file(identity_file, conf['appliance_url']),
                {
                    **({"id": authn_login} if authn_login is not None else {}),
                    **({"api_key": authn_api_key} if authn_api_key is not None else {})
                }
            )

            if 'id' not in identity:
                # For authn-cert SPIFFE mode the workload identity is derived
                # from the SPIFFE URI in the client certificate; no login needed.
                # For authn-jwt token-app-property mode the workload identity is
                # derived from a JWT claim configured in the Conjur policy; no login needed.
                if authn_type == 'authn-cert' and cert_mode == 'spiffe':
                    pass
                elif authn_type == 'jwt' and jwt_mode == 'token-app-property':
                    pass
                else:
                    raise AnsibleError(
                        """Configuration must define options `conjur_authn_login`.
                    This config can be set by any of the following methods, listed in order of priority:
                    - Ansible variable `conjur_authn_login`, set either in the parent playbook or passed via --extra-vars
                    - Environment variable `CONJUR_AUTHN_LOGIN`
                    - An identity file with the field `login`"""
                    )

        cert_file = None
        if 'cert_file' in conf:
            display.vvv(f"Using cert file path {conf['cert_file']}")
            cert_file = conf['cert_file']

        try:
            token = None
            if 'authn_token_file' not in conf:
                display.vvv(f"Using auth_type as {authn_type}")
                if authn_type == 'aws':
                    token = _fetch_conjur_iam_session_token(
                        appliance_url=conf['appliance_url'],
                        account=conf['account'],
                        host_id=identity['id'],
                        service_id=service_id,
                        validate_certs=validate_certs,
                        cert_file=cert_file
                    )
                elif authn_type == "azure":
                    token = _fetch_conjur_azure_token(
                        appliance_url=conf['appliance_url'],
                        account=conf['account'],
                        host_id=identity['id'],
                        service_id=service_id,
                        validate_certs=validate_certs,
                        cert_file=cert_file,
                        client_id=azure_client_id
                    )
                elif authn_type == "gcp":
                    token = _fetch_conjur_gcp_identity_token(
                        appliance_url=conf['appliance_url'],
                        account=conf['account'],
                        host_id=identity['id'],
                        validate_certs=validate_certs,
                        cert_file=cert_file,
                    )
                elif authn_type == "jwt":
                    if jwt_token_file and not jwt_token:
                        jwt_token = _read_jwt_token_file(jwt_token_file)
                    token = _fetch_conjur_jwt_token(
                        appliance_url=conf['appliance_url'],
                        account=conf['account'],
                        service_id=service_id,
                        # token-app-property mode: omit host_id from URL; identity
                        # is resolved from the JWT claim mapping in Conjur policy.
                        host_id=identity.get('id') if jwt_mode != 'token-app-property' else None,
                        jwt_token=jwt_token,
                        validate_certs=validate_certs,
                        cert_file=cert_file,
                    )
                elif authn_type == "authn-cert":
                    token = _fetch_conjur_cert_token(
                        appliance_url=conf['appliance_url'],
                        account=conf['account'],
                        service_id=service_id,
                        # spiffe mode: identity comes from the certificate, not the URL
                        host_id=identity.get('id') if cert_mode != 'spiffe' else None,
                        client_cert_file=client_cert_file,
                        client_key_file=client_key_file,
                        ca_cert_file=cert_file,
                        validate_certs=validate_certs
                    )
                else:
                    token = _fetch_conjur_token(
                        conf['appliance_url'],
                        conf['account'],
                        identity['id'],
                        identity['api_key'],
                        validate_certs,
                        cert_file
                    )
            else:
                if not os.path.exists(conf['authn_token_file']):
                    raise AnsibleError(f"Conjur authn token file `{conf['authn_token_file']}` was not found on the host")
                with open(conf['authn_token_file'], 'rb') as file:
                    token = file.read()

            conjur_variable = _fetch_conjur_variable(
                terms[0],
                token,
                conf['appliance_url'],
                conf['account'],
                validate_certs,
                cert_file,
                retry_interval
            )
        finally:
            if isinstance(token, bytes):
                token = b"\x00" * len(token)
            else:
                token = None
            jwt_token = None

            if temp_cert_file:
                try:
                    if os.path.exists(temp_cert_file.name):
                        os.unlink(temp_cert_file.name)
                except (OSError, PermissionError) as err:
                    raise AnsibleError(f"Failed to delete temporary certificate file `{temp_cert_file.name}`: {str(err)}") from err

        if as_file:
            return _store_secret_in_file(conjur_variable)

        return conjur_variable

    def get_var_value(self, key):
        try:
            variable_value = self.get_option(key)
        except KeyError as err:
            raise AnsibleError(f"{key} was not defined in configuration") from err

        return variable_value



#!/usr/bin/env python3
"""
Script para ejecutar todos los playbooks de Ansible y generar un informe
con los resultados de éxito y fallos.
"""

import os
import re
import subprocess
import sys
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple


class PlaybookRunner:
    def __init__(self, playbooks_dir: str = "playbooks", logs_dir: str = "playbooks_logs"):
        self.playbooks_dir = Path(playbooks_dir)
        self.logs_dir = Path(logs_dir)
        self.results: List[Dict] = []
        
        # Crear directorio de logs si no existe
        self.logs_dir.mkdir(exist_ok=True)
        
        # Códigos de salida de ansible-playbook y sus significados
        self.exit_codes = {
            0: "Éxito - El playbook se ejecutó correctamente",
            1: "Error general - Ocurrió un error durante la ejecución",
            2: "Tareas fallidas - Una o más tareas fallaron en uno o más hosts",
            3: "Hosts inalcanzables - Uno o más hosts no fueron alcanzables",
            4: "Error de sintaxis - Problemas con la sintaxis del playbook o opciones",
            5: "Mal uso - Uso incorrecto de ansible-playbook",
            99: "Error de usuario - Error causado por el usuario",
            250: "Error inesperado - Error inesperado en ansible-playbook"
        }
        
    def find_playbooks(self) -> List[Path]:
        """Encuentra todos los archivos playbook (.yml y .yaml) en el directorio."""
        playbooks = []
        if not self.playbooks_dir.exists():
            print(f"Error: El directorio {self.playbooks_dir} no existe")
            sys.exit(1)
            
        for ext in ['*.yml', '*.yaml']:
            playbooks.extend(self.playbooks_dir.glob(ext))
        
        # Excluir archivos que no son playbooks (como credentials.template)
        playbooks = [p for p in playbooks if p.is_file() and 
                    not p.name.endswith('.template')]
        
        return sorted(playbooks)
    
    def get_exit_code_meaning(self, exit_code: int) -> str:
        """
        Obtiene el significado de un código de salida de ansible-playbook.
        
        Args:
            exit_code: Código de salida
            
        Returns:
            Descripción del código de salida
        """
        return self.exit_codes.get(exit_code, f"Código desconocido: {exit_code}")
    
    def check_missing_deps(self, playbook_path: Path) -> List[str]:
        """Detecta vars files referenciados en el playbook que no existen en disco."""
        missing = []
        try:
            with open(playbook_path, encoding='utf-8') as f:
                content = f.read()
        except Exception:
            return missing

        # Buscar include_vars con file: o vars_files
        patterns = [
            r"file:\s+['\"]?([^\s'\"]+\.ya?ml)['\"]?",
            r"vars_files:\s*\n\s+-\s+['\"]?([^\s'\"]+\.ya?ml)['\"]?",
        ]
        for pattern in patterns:
            for match in re.finditer(pattern, content):
                dep_file = match.group(1)
                # Resolver relativo al directorio del playbook
                dep_path = playbook_path.parent / dep_file
                if not dep_path.exists():
                    missing.append(dep_file)
        return missing

    def run_playbook(self, playbook_path: Path) -> Tuple[bool, str, str, int]:
        """
        Ejecuta un playbook y retorna (éxito, stdout, stderr, exit_code).

        Args:
            playbook_path: Ruta al playbook

        Returns:
            Tupla con (éxito, stdout, stderr, exit_code)
        """
        print(f"\n{'='*80}")
        print(f"Ejecutando: {playbook_path.name}")
        print(f"{'='*80}")

        # Pre-check: detectar dependencias faltantes antes de correr
        missing_deps = self.check_missing_deps(playbook_path)
        if missing_deps:
            msg = f"Dependencias faltantes: {', '.join(missing_deps)}"
            print(f"⚠ {playbook_path.name} - SKIP ({msg})")
            return None, "", msg, -2
        
        try:
            # Ejecutar ansible-playbook con captura de salida
            result = subprocess.run(
                ['ansible-playbook', '-i', 'localhost,', str(playbook_path)],
                capture_output=True,
                text=True,
                timeout=600,  # Timeout de 10 minutos por playbook
                cwd=self.playbooks_dir.parent
            )
            
            success = result.returncode == 0
            stdout = result.stdout
            stderr = result.stderr
            exit_code = result.returncode

            if success:
                print(f"✓ {playbook_path.name} - ÉXITO")
            else:
                exit_meaning = self.get_exit_code_meaning(exit_code)
                print(f"✗ {playbook_path.name} - FALLÓ")
                print(f"  Código de salida: {exit_code} - {exit_meaning}")
                print(f"  Log completo se guardará en: {self.logs_dir}/")

            return success, stdout, stderr, exit_code
            
        except subprocess.TimeoutExpired:
            error_msg = f"Timeout: El playbook tardó más de 10 minutos en ejecutarse"
            print(f"✗ {playbook_path.name} - TIMEOUT")
            return False, "", error_msg, -1
            
        except FileNotFoundError:
            error_msg = "Error: ansible-playbook no se encontró. ¿Está instalado Ansible?"
            print(f"✗ {playbook_path.name} - ERROR: {error_msg}")
            return False, "", error_msg, -1
            
        except Exception as e:
            error_msg = f"Error inesperado: {str(e)}"
            print(f"✗ {playbook_path.name} - ERROR: {error_msg}")
            return False, "", error_msg, -1
    
    def save_logs(self, playbook_name: str, stdout: str, stderr: str, exit_code: int) -> str:
        """
        Guarda los logs completos de un playbook fallido en un archivo.
        
        Args:
            playbook_name: Nombre del playbook
            stdout: Salida estándar
            stderr: Salida de error
            exit_code: Código de salida
            
        Returns:
            Ruta al archivo de log guardado
        """
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        log_filename = f"{playbook_name}_{timestamp}.log"
        log_path = self.logs_dir / log_filename
        
        with open(log_path, 'w', encoding='utf-8') as f:
            f.write("=" * 80 + "\n")
            f.write(f"LOG DE EJECUCIÓN: {playbook_name}\n")
            f.write("=" * 80 + "\n")
            f.write(f"Fecha: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")
            f.write(f"Código de salida: {exit_code} - {self.get_exit_code_meaning(exit_code)}\n")
            f.write("\n" + "=" * 80 + "\n")
            f.write("SALIDA ESTÁNDAR (STDOUT)\n")
            f.write("=" * 80 + "\n")
            f.write(stdout)
            f.write("\n\n" + "=" * 80 + "\n")
            f.write("SALIDA DE ERROR (STDERR)\n")
            f.write("=" * 80 + "\n")
            f.write(stderr)
            f.write("\n" + "=" * 80 + "\n")
        
        return str(log_path)
    
    def extract_error_reason(self, stdout: str, stderr: str) -> str:
        """
        Extrae la razón del fallo de la salida de ansible-playbook.
        
        Args:
            stdout: Salida estándar
            stderr: Salida de error
            
        Returns:
            Razón del fallo extraída
        """
        # Combinar stdout y stderr para buscar errores
        output = stdout + "\n" + stderr
        
        # Detectar errores específicos conocidos
        if "requires Python 3.10" in output or "Python 3.10 at minimum" in output:
            python_version = sys.version.split()[0]
            return f"❌ ERROR DE VERSIÓN DE PYTHON: Se requiere Python 3.10+, pero estás usando {python_version}. Actualiza Python o usa un entorno virtual con Python 3.10+."
        
        if "Could not find or access" in output:
            match = re.search(r"Could not find or access '([^']+)'", output)
            if match:
                return f"❌ ARCHIVO NO ENCONTRADO: No se pudo encontrar el archivo '{match.group(1)}'"
        
        if "authentication failed" in output.lower() or "unauthorized" in output.lower():
            return "❌ ERROR DE AUTENTICACIÓN: Credenciales inválidas o expiradas. Verifica tu API key de Meraki."
        
        if "network_id" in output.lower() and ("required" in output.lower() or "missing" in output.lower()):
            return "❌ VARIABLE FALTANTE: Se requiere 'network_id' pero no está definida en el playbook."
        
        if "org_id" in output.lower() and ("required" in output.lower() or "missing" in output.lower()):
            return "❌ VARIABLE FALTANTE: Se requiere 'org_id' pero no está definida en el playbook."
        
        # Buscar patrones comunes de error
        error_patterns = [
            r'ERROR! (.+?)(?:\n|$)',
            r'fatal: (.+?)(?:\n|$)',
            r'failed: (.+?)(?:\n|$)',
            r'Error: (.+?)(?:\n|$)',
            r'Exception: (.+?)(?:\n|$)',
        ]
        
        reasons = []
        for pattern in error_patterns:
            matches = re.findall(pattern, output, re.IGNORECASE | re.MULTILINE)
            if matches:
                reasons.extend(matches[:2])  # Tomar máximo 2 coincidencias por patrón
        
        if reasons:
            # Unir las razones y limitar la longitud
            reason = "; ".join(reasons[:3])
            if len(reason) > 200:
                reason = reason[:200] + "..."
            return reason
        
        # Si no se encontró un patrón específico, buscar las últimas líneas de error
        lines = output.split('\n')
        error_lines = [line for line in lines if any(keyword in line.lower() 
                    for keyword in ['error', 'fatal', 'failed', 'exception'])]
        
        if error_lines:
            # Tomar las últimas líneas de error
            last_errors = error_lines[-3:]
            reason = " | ".join(last_errors)
            if len(reason) > 200:
                reason = reason[:200] + "..."
            return reason
        
        # Si no se encontró nada específico, retornar un mensaje genérico
        if stderr:
            return stderr[:200] if len(stderr) <= 200 else stderr[:200] + "..."
        
        return "Razón desconocida (revisar logs)"
    
    def categorize_error(self, error_reason: str) -> str:
        """
        Categoriza un error en un tipo específico.
        
        Args:
            error_reason: Razón del error
            
        Returns:
            Categoría del error
        """
        if "VERSIÓN DE PYTHON" in error_reason:
            return "Versión de Python incompatible"
        elif "ARCHIVO NO ENCONTRADO" in error_reason:
            return "Archivo faltante"
        elif "AUTENTICACIÓN" in error_reason:
            return "Error de autenticación"
        elif "VARIABLE FALTANTE" in error_reason:
            return "Variable faltante"
        else:
            return "Otro error"
    
    def check_python_version(self) -> None:
        """Verifica la versión de Python y muestra advertencias si es necesario."""
        python_version = sys.version_info
        current_version = f"{python_version.major}.{python_version.minor}.{python_version.micro}"
        
        print(f"Python detectado: {current_version}")
        
        if python_version < (3, 10):
            print(f"\n⚠️  ADVERTENCIA: Estás usando Python {current_version}")
            print("   La librería Meraki SDK requiere Python 3.10 o superior.")
            print("   Muchos playbooks pueden fallar por esta razón.")
            print("   Considera actualizar Python o usar un entorno virtual con Python 3.10+")
            print("")
        
    def run_all(self) -> None:
        """Ejecuta todos los playbooks y almacena los resultados."""
        playbooks = self.find_playbooks()
        
        if not playbooks:
            print(f"No se encontraron playbooks en {self.playbooks_dir}")
            return
        
        print(f"\n{'='*80}")
        print(f"Encontrados {len(playbooks)} playbooks para ejecutar")
        print(f"Logs de playbooks fallidos se guardarán en: {self.logs_dir}/")
        print(f"{'='*80}\n")
        
        # Verificar versión de Python
        self.check_python_version()
        
        for i, playbook in enumerate(playbooks, 1):
            print(f"\n[{i}/{len(playbooks)}] Procesando: {playbook.name}")
            
            success, stdout, stderr, exit_code = self.run_playbook(playbook)

            # None = skipped por dependencias faltantes
            if success is None:
                self.results.append({
                    'playbook': playbook.name,
                    'path': str(playbook),
                    'success': None,
                    'exit_code': -2,
                    'exit_meaning': 'Skipped',
                    'error_reason': stderr,
                    'error_category': 'Dependencia faltante',
                    'stdout': '',
                    'stderr': stderr,
                    'log_file': None,
                })
                continue

            error_reason = ""
            exit_meaning = ""
            error_category = ""
            if not success:
                error_reason = self.extract_error_reason(stdout, stderr)
                exit_meaning = self.get_exit_code_meaning(exit_code)
                error_category = self.categorize_error(error_reason)

            log_file = None
            if not success:
                log_file = self.save_logs(playbook.name, stdout, stderr, exit_code)

            self.results.append({
                'playbook': playbook.name,
                'path': str(playbook),
                'success': success,
                'exit_code': exit_code,
                'exit_meaning': exit_meaning,
                'error_reason': error_reason,
                'error_category': error_category,
                'stdout': stdout,
                'stderr': stderr,
                'log_file': log_file,
            })
    
    def generate_report(self, output_file: str = None) -> str:
        """
        Genera un informe con los resultados de la ejecución.
        
        Args:
            output_file: Archivo donde guardar el informe (opcional)
            
        Returns:
            Contenido del informe
        """
        total = len(self.results)
        successful = sum(1 for r in self.results if r['success'] is True)
        skipped = sum(1 for r in self.results if r['success'] is None)
        failed = total - successful - skipped
        
        report_lines = []
        report_lines.append("=" * 80)
        report_lines.append("INFORME DE EJECUCIÓN DE PLAYBOOKS")
        report_lines.append("=" * 80)
        report_lines.append(f"Fecha: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
        report_lines.append(f"Total de playbooks: {total}")
        report_lines.append(f"Éxitos : {successful} ({successful/total*100:.1f}%)" if total > 0 else "Éxitos: 0")
        report_lines.append(f"Fallos : {failed} ({failed/total*100:.1f}%)" if total > 0 else "Fallos: 0")
        report_lines.append(f"Saltados: {skipped} (dependencias faltantes)" if skipped > 0 else "Saltados: 0")
        if failed > 0:
            report_lines.append(f"Logs de playbooks fallidos guardados en: {self.logs_dir}/")
        report_lines.append("")

        if successful > 0:
            report_lines.append("-" * 80)
            report_lines.append("PLAYBOOKS EXITOSOS")
            report_lines.append("-" * 80)
            for result in self.results:
                if result['success'] is True:
                    report_lines.append(f"✓ {result['playbook']}")
            report_lines.append("")

        if skipped > 0:
            report_lines.append("-" * 80)
            report_lines.append("PLAYBOOKS SALTADOS (dependencia faltante)")
            report_lines.append("-" * 80)
            for result in self.results:
                if result['success'] is None:
                    report_lines.append(f"⚠ {result['playbook']}")
                    report_lines.append(f"  {result['error_reason']}")
            report_lines.append("")

        if failed > 0:
            # Agrupar errores por categoría
            errors_by_category = defaultdict(list)
            for result in self.results:
                if result['success'] is False:
                    category = result.get('error_category', 'Otro error')
                    errors_by_category[category].append(result)
            
            report_lines.append("-" * 80)
            report_lines.append("RESUMEN DE ERRORES POR CATEGORÍA")
            report_lines.append("-" * 80)
            for category, errors in sorted(errors_by_category.items(), key=lambda x: len(x[1]), reverse=True):
                report_lines.append(f"{category}: {len(errors)} playbook(s)")
            report_lines.append("")
            
            report_lines.append("-" * 80)
            report_lines.append("PLAYBOOKS FALLIDOS (DETALLES)")
            report_lines.append("-" * 80)
            
            # Mostrar primero los errores más comunes
            for category in sorted(errors_by_category.keys(), key=lambda x: len(errors_by_category[x]), reverse=True):
                if len(errors_by_category[category]) > 0:
                    report_lines.append("")
                    report_lines.append(f"📋 {category.upper()} ({len(errors_by_category[category])} playbook(s))")
                    report_lines.append("-" * 80)
                    
                    for result in errors_by_category[category]:
                        report_lines.append(f"✗ {result['playbook']}")
                        if result['exit_code'] != -1:
                            report_lines.append(f"  Código de salida: {result['exit_code']} - {result['exit_meaning']}")
                        if result['error_reason']:
                            report_lines.append(f"  Razón: {result['error_reason']}")
                        if result.get('log_file'):
                            report_lines.append(f"  Log completo: {result['log_file']}")
                        report_lines.append("")
        
        report_content = "\n".join(report_lines)
        
        if output_file:
            with open(output_file, 'w', encoding='utf-8') as f:
                f.write(report_content)
            print(f"\n✓ Informe guardado en: {output_file}")
        
        return report_content


def main():
    """Función principal."""
    import argparse
    
    parser = argparse.ArgumentParser(
        description='Ejecuta todos los playbooks de Ansible y genera un informe'
    )
    parser.add_argument(
        '--playbooks-dir',
        default='playbooks',
        help='Directorio donde están los playbooks (default: playbooks)'
    )
    parser.add_argument(
        '--output',
        '-o',
        default='playbooks_report.txt',
        help='Archivo de salida para el informe (default: playbooks_report.txt)'
    )
    parser.add_argument(
        '--no-report-file',
        action='store_true',
        help='No guardar el informe en un archivo, solo mostrarlo en consola'
    )
    parser.add_argument(
        '--logs-dir',
        default='playbooks_logs',
        help='Directorio donde guardar los logs de playbooks fallidos (default: playbooks_logs)'
    )
    
    args = parser.parse_args()
    
    runner = PlaybookRunner(playbooks_dir=args.playbooks_dir, logs_dir=args.logs_dir)
    
    try:
        runner.run_all()
        
        if args.no_report_file:
            report = runner.generate_report()
            print("\n" + report)
        else:
            report = runner.generate_report(output_file=args.output)
            print("\n" + report)
        
        # Retornar código de salida apropiado
        failed_count = sum(1 for r in runner.results if r['success'] is False)
        sys.exit(0 if failed_count == 0 else 1)
        
    except KeyboardInterrupt:
        print("\n\nEjecución interrumpida por el usuario")
        sys.exit(130)
    except Exception as e:
        print(f"\nError fatal: {e}")
        import traceback
        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    main()




# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: _latest_version
short_description: "[INTERNAL] Get latest version from a list of versions"
version_added: 1.4.0
author:
  - Felix Fontein (@felixfontein)
description:
  - B(This is an internal tool and must only be used from roles in this collection!) If you use it from outside this collection,
    be warned that its behavior can change and it can be removed at any time, even in bugfix releases!
  - Given a list of version numbers, returns the largest of them.
options:
  _input:
    description:
      - A list of strings. Every string must be a version number.
    type: list
    elements: string
    required: true
"""

EXAMPLES = r"""
---
- name: Print latest version
  ansible.builtin.debug:
    msg: "{{ versions | community.sops._latest_version }}"
  vars:
    versions:
      - 1.0.0
      - 1.0.0rc1
      - 1.1.0
"""

RETURN = r"""
_value:
  description:
    - The latest version from the input.
    - Returns the empty string if the input was empty.
  type: string
"""

from ansible.module_utils.compat.version import LooseVersion


def pick_latest_version(version_list):
    '''Pick latest version from a list of versions.'''
    # Remove all prereleases (versions with '+' or '-' in them)
    version_list = [v for v in version_list if '-' not in v and '+' not in v]
    if not version_list:
        return ''
    return max(version_list, key=LooseVersion)


class FilterModule:
    '''Helper filters.'''
    def filters(self):
        return {
            '_latest_version': pick_latest_version,
        }



# Copyright (c) Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import typing as t
from collections.abc import Mapping

import tomllib
from ansible.errors import AnsibleFilterError


def from_toml(value: t.Any) -> Mapping:
    if not isinstance(value, str):
        raise AnsibleFilterError("from_toml only accepts strings.")
    return tomllib.loads(value)


class FilterModule:
    def filters(self):
        return {
            "from_toml": from_toml,
        }



# Copyright (c) 2024 Red Hat
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
import json
import os
import shlex
import tempfile
import time
import hashlib
import sys

from ansible_collections.containers.podman.plugins.module_utils.podman.common import run_podman_command

__metaclass__ = type


class PodmanImageError(Exception):
    """Custom exception for Podman image operations."""


class ImageRepository:
    """Parse and manage image repository information."""

    def __init__(self, name, tag="latest"):
        self.original_name = name
        self.name, self.parsed_tag = self._parse_repository_tag(name)
        self.tag = self.parsed_tag or tag
        self.delimiter = "@" if self.tag.startswith("sha256:") else ":"
        self.full_name = f"{self.name}{self.delimiter}{self.tag}"

    @staticmethod
    def _parse_repository_tag(repo_name):
        """Parse repository name and tag/digest."""
        parts = repo_name.rsplit("@", 1)
        if len(parts) == 2:
            return tuple(parts)
        parts = repo_name.rsplit(":", 1)
        if len(parts) == 2 and "/" not in parts[1]:
            return tuple(parts)
        return repo_name, None


class ContainerFileProcessor:
    """Handle Containerfile/Dockerfile processing and hashing."""

    def __init__(self, build_config, path=None):
        self.build_config = build_config or {}
        self.path = path

    def get_containerfile_contents(self):
        """Get Containerfile contents from various sources."""
        build_file_arg = self.build_config.get("file")
        containerfile_contents = self.build_config.get("container_file")

        container_filename = None
        if build_file_arg:
            container_filename = build_file_arg
        elif self.path and not build_file_arg:
            container_filename = self._find_containerfile_from_context()

        if not containerfile_contents and container_filename and os.access(container_filename, os.R_OK):
            with open(container_filename, "r", encoding='utf-8') as f:
                containerfile_contents = f.read()

        return containerfile_contents

    def _find_containerfile_from_context(self):
        """Find Containerfile/Dockerfile in build context."""
        if not self.path:
            return None

        for filename in ["Containerfile", "Dockerfile"]:
            full_path = os.path.join(self.path, filename)
            if os.path.exists(full_path):
                return full_path
        return None

    def hash_containerfile_contents(self, contents):
        """Generate SHA256 hash of Containerfile contents."""
        if not contents:
            return None

        if sys.version_info < (3, 9):
            return hashlib.sha256(contents.encode()).hexdigest()
        return hashlib.sha256(contents.encode(), usedforsecurity=False).hexdigest()


class PodmanImageInspector:
    """Handle image inspection operations."""

    def __init__(self, module, executable):
        self.module = module
        self.executable = executable

    def inspect_image(self, image_name):
        """Inspect an image and return its data."""
        args = ["inspect", image_name, "--format", "json"]
        rc, out, unused = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            return None

        try:
            image_data = json.loads(out)
            return image_data if image_data else None
        except json.JSONDecodeError:
            self.module.fail_json(msg=f"Failed to parse JSON output from podman inspect: {out}")

    def image_exists(self, image_name):
        """Check if an image exists."""
        rc, out, err = run_podman_command(
            self.module, self.executable, ["image", "exists", image_name], ignore_errors=True
        )
        return rc == 0

    def list_images(self, image_name):
        """List images matching the given name."""
        args = ["image", "ls", image_name, "--format", "json"]
        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            return []

        try:
            return json.loads(out) or []
        except json.JSONDecodeError:
            self.module.fail_json(msg=f"Failed to parse JSON output from podman image ls: {out}")


class PodmanImageBuilder:
    """Handle image building operations."""

    def __init__(self, module, executable, auth_config=None):
        self.module = module
        self.executable = executable
        self.auth_config = auth_config or {}

    def build_image(self, image_name, build_config, path=None, containerfile_hash=None, platform=None, arch=None):
        """Build an image with the given configuration."""
        args = self._construct_build_args(image_name, build_config, path, containerfile_hash, platform, arch)

        # Handle inline container file
        temp_file_path = None
        if build_config.get("container_file"):
            temp_file_path = self._create_temp_containerfile(build_config["container_file"], path)
            args.extend(["--file", temp_file_path])

        try:
            # Return the command that will be executed for podman_actions tracking
            podman_command = " ".join([self.executable] + args)
            rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

            if rc != 0:
                self.module.fail_json(msg=f"Failed to build image {image_name}: {out} {err}")

            # Extract image ID from output
            last_id = self._extract_image_id_from_output(out)
            return last_id, out + err, podman_command

        finally:
            if temp_file_path and os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def _construct_build_args(self, image_name, build_config, path, containerfile_hash, platform=None, arch=None):
        """Construct build command arguments."""
        args = ["build", "-t", image_name]

        # Add authentication
        self._add_auth_args(args)

        # Add arch/platform for cross-platform builds
        if arch:
            args.extend(["--arch", arch])

        if platform:
            args.extend(["--platform", platform])

        # Add build-specific arguments
        if build_config.get("force_rm"):
            args.append("--force-rm")

        if build_config.get("format"):
            args.extend(["--format", build_config["format"]])

        if not build_config.get("cache", True):
            args.append("--no-cache")

        if build_config.get("rm", True):
            args.append("--rm")

        # Add annotations
        annotations = build_config.get("annotation") or {}
        for key, value in annotations.items():
            args.extend(["--annotation", f"{key}={value}"])

        # Add containerfile hash as label
        if containerfile_hash:
            args.extend(["--label", f"containerfile.hash={containerfile_hash}"])

        # Add volumes
        volumes = build_config.get("volume") or []
        for volume in volumes:
            if volume:
                args.extend(["--volume", volume])

        # Add build file
        if build_config.get("file"):
            args.extend(["--file", build_config["file"]])

        # Add target
        if build_config.get("target"):
            args.extend(["--target", build_config["target"]])

        # Add extra args
        extra_args = build_config.get("extra_args")
        if extra_args:
            args.extend(shlex.split(extra_args))

        # Add build context path
        if path:
            args.append(path)

        return args

    def _add_auth_args(self, args):
        """Add authentication arguments to command."""
        if self.auth_config.get("validate_certs") is not None:
            if self.auth_config["validate_certs"]:
                args.append("--tls-verify")
            else:
                args.append("--tls-verify=false")

        if self.auth_config.get("ca_cert_dir"):
            args.extend(["--cert-dir", self.auth_config["ca_cert_dir"]])

        if self.auth_config.get("auth_file"):
            args.extend(["--authfile", self.auth_config["auth_file"]])

        if self.auth_config.get("username") and self.auth_config.get("password"):
            cred_string = f"{self.auth_config['username']}:{self.auth_config['password']}"
            args.extend(["--creds", cred_string])

    def _create_temp_containerfile(self, content, path):
        """Create a temporary Containerfile with the given content."""
        if path:
            temp_path = os.path.join(path, f"Containerfile.generated_by_ansible_{time.time()}")
        else:
            temp_dir = tempfile.gettempdir()
            temp_path = os.path.join(temp_dir, f"Containerfile.generated_by_ansible_{time.time()}")

        with open(temp_path, "w", encoding='utf-8') as f:
            f.write(content)

        return temp_path

    def _extract_image_id_from_output(self, output):
        """Extract image ID from build output."""
        for line in output.splitlines():
            if line.startswith("-->"):
                parts = line.rsplit(" ", 1)
                if len(parts) > 1:
                    return parts[1]

        # Fallback to last line if no --> found
        lines = output.splitlines()
        return lines[-1] if lines else ""


class PodmanImagePuller:
    """Handle image pulling operations."""

    def __init__(self, module, executable, auth_config=None):
        self.module = module
        self.executable = executable
        self.auth_config = auth_config or {}

    def pull_image(self, image_name, arch=None, platform=None, pull_extra_args=None):
        """Pull an image from a registry."""
        args = ["pull", image_name]

        if platform:
            args.extend(["--platform", platform])
        elif arch:
            args.extend(["--arch", arch])

        self._add_auth_args(args)

        if pull_extra_args:
            args.extend(shlex.split(pull_extra_args))

        # Return the command that will be executed for podman_actions tracking
        podman_command = " ".join([self.executable] + args)
        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            self.module.fail_json(msg=f"Failed to pull image {image_name}", stderr=err, stdout=out)

        return out.strip(), podman_command

    def _add_auth_args(self, args):
        """Add authentication arguments to command."""
        if self.auth_config.get("validate_certs") is not None:
            if self.auth_config["validate_certs"]:
                args.append("--tls-verify")
            else:
                args.append("--tls-verify=false")

        if self.auth_config.get("ca_cert_dir"):
            args.extend(["--cert-dir", self.auth_config["ca_cert_dir"]])

        if self.auth_config.get("auth_file"):
            args.extend(["--authfile", self.auth_config["auth_file"]])

        if self.auth_config.get("username") and self.auth_config.get("password"):
            cred_string = f"{self.auth_config['username']}:{self.auth_config['password']}"
            args.extend(["--creds", cred_string])


class PodmanImagePusher:
    """Handle image pushing operations."""

    def __init__(self, module, executable, auth_config=None):
        self.module = module
        self.executable = executable
        self.auth_config = auth_config or {}

    def push_image(self, image_name, push_config):
        """Push an image to a registry."""
        transport = (push_config or {}).get("transport")

        # Special handling for scp transport which uses 'podman image scp'
        if transport == "scp":
            args = []

            # Allow passing global --ssh options to podman
            ssh_opts = push_config.get("ssh") if push_config else None
            if ssh_opts:
                args.extend(["--ssh", ssh_opts])

            args.extend(["image", "scp"])

            # Extra args (e.g., --quiet) if provided
            if push_config.get("extra_args"):
                args.extend(shlex.split(push_config["extra_args"]))

            # Source image (local)
            args.append(image_name)

            # Destination host spec
            dest = push_config.get("dest")
            if not dest:
                self.module.fail_json(msg="When using transport 'scp', push_args.dest must be provided")

            # If user did not include '::' in dest, append it to copy into remote storage with same name
            dest_spec = dest if "::" in dest else f"{dest}::"
            args.append(dest_spec)

            action = " ".join(args)
            rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)
            if rc != 0:
                self.module.fail_json(
                    msg=f"Failed to scp image {image_name} to {dest}", stdout=out, stderr=err, actions=[action]
                )
            return out + err, action

        # Default push behavior for all other transports
        args = ["push"]

        self._add_auth_args(args)
        self._add_push_args(args, push_config)

        args.append(image_name)

        # Build destination
        dest_string = self._build_destination(image_name, push_config)
        args.append(dest_string)
        action = " ".join(args)

        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            self.module.fail_json(msg=f"Failed to push image {image_name}", stdout=out, stderr=err, actions=[action])

        return out + err, action

    def _add_auth_args(self, args):
        """Add authentication arguments to command."""
        if self.auth_config.get("validate_certs") is not None:
            if self.auth_config["validate_certs"]:
                args.append("--tls-verify")
            else:
                args.append("--tls-verify=false")

        if self.auth_config.get("ca_cert_dir"):
            args.extend(["--cert-dir", self.auth_config["ca_cert_dir"]])

        if self.auth_config.get("auth_file"):
            args.extend(["--authfile", self.auth_config["auth_file"]])

        if self.auth_config.get("username") and self.auth_config.get("password"):
            cred_string = f"{self.auth_config['username']}:{self.auth_config['password']}"
            args.extend(["--creds", cred_string])

    def _add_push_args(self, args, push_config):
        """Add push-specific arguments."""
        if push_config.get("compress"):
            args.append("--compress")

        if push_config.get("format"):
            args.extend(["--format", push_config["format"]])

        if push_config.get("remove_signatures"):
            args.append("--remove-signatures")

        if push_config.get("sign_by"):
            args.extend(["--sign-by", push_config["sign_by"]])

        if push_config.get("extra_args"):
            args.extend(shlex.split(push_config["extra_args"]))

    def _build_destination(self, image_name, push_config):
        """Build the destination string for push."""
        dest = push_config.get("dest") or image_name
        transport = push_config.get("transport")

        if not transport:
            # Handle simple registry destination
            if ":" not in dest and "@" not in dest and len(dest.rstrip("/").split("/")) == 2:
                return f"{dest.rstrip('/')}/{image_name}"
            if "/" not in dest and "@" not in dest and "docker-daemon" not in dest:
                self.module.fail_json(
                    msg="Destination must be a full URL or path to a directory with image name and tag."
                )
            return dest

        # Handle transport-specific destinations
        name_parts = image_name.split("/")
        image_base_name = name_parts[-1].split(":")[0]

        if transport == "docker":
            return f"{transport}://{dest}"
        if transport == "ostree":
            return f"{transport}:{image_base_name}@{dest}"
        if transport == "docker-daemon":
            if ":" not in dest:
                return f"{transport}:{dest}:latest"
            return f"{transport}:{dest}"
        return f"{transport}:{dest}"


class PodmanImageRemover:
    """Handle image removal operations."""

    def __init__(self, module, executable):
        self.module = module
        self.executable = executable

    def remove_image(self, image_name, force=False):
        """Remove an image by name."""
        args = ["rmi", image_name]
        if force:
            args.append("--force")

        # Return the command that will be executed for podman_actions tracking
        podman_command = " ".join([self.executable] + args)
        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            self.module.fail_json(msg=f"Failed to remove image {image_name}: {err}")

        return out, podman_command

    def remove_image_by_id(self, image_id, force=False):
        """Remove an image by ID."""
        args = ["rmi", image_id]
        if force:
            args.append("--force")

        # Return the command that will be executed for podman_actions tracking
        podman_command = " ".join([self.executable] + args)
        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            self.module.fail_json(msg=f"Failed to remove image with ID {image_id}: {err}")

        return out, podman_command


class PodmanImageManager:
    """Main image management class that orchestrates all operations."""

    def __init__(self, module):
        self.module = module
        self.params = module.params
        self.executable = module.get_bin_path(self.params.get("executable", "podman"), required=True)

        # Initialize repository info
        self.repository = ImageRepository(self.params["name"], self.params.get("tag", "latest"))

        # Initialize components
        self.inspector = PodmanImageInspector(self.module, self.executable)

        auth_config = self._build_auth_config()
        self.builder = PodmanImageBuilder(self.module, self.executable, auth_config)
        self.puller = PodmanImagePuller(self.module, self.executable, auth_config)
        self.pusher = PodmanImagePusher(self.module, self.executable, auth_config)
        self.remover = PodmanImageRemover(self.module, self.executable)

        # Initialize containerfile processor
        self.containerfile_processor = ContainerFileProcessor(self.params.get("build", {}), self.params.get("path"))

        # Results tracking
        self.results = {
            "changed": False,
            "actions": [],
            "podman_actions": [],
            "image": {},
            "stdout": "",
        }

    def _build_auth_config(self):
        """Build authentication configuration."""
        return {
            "validate_certs": self.params.get("validate_certs"),
            "ca_cert_dir": self.params.get("ca_cert_dir"),
            "auth_file": self.params.get("auth_file"),
            "username": self.params.get("username"),
            "password": self.params.get("password"),
        }

    def find_image(self, image_name=None):
        """Find an image and return its information."""
        image_name = image_name or self.repository.full_name

        if not self.inspector.image_exists(image_name):
            return None

        images = self.inspector.list_images(image_name)
        if not images:
            return None

        # Check architecture if specified
        arch = self.params.get("arch")
        platform = self.params.get("platform")
        if arch:
            inspect_data = self.inspector.inspect_image(image_name)
            if inspect_data and inspect_data[0].get("Architecture") != arch:
                return None
        elif platform:
            # Platform format: os/arch or os/arch/variant (e.g. linux/amd64)
            inspect_data = self.inspector.inspect_image(image_name)
            if inspect_data:
                platform_parts = platform.split("/")
                required_os = platform_parts[0] if len(platform_parts) >= 1 else None
                required_arch = platform_parts[1] if len(platform_parts) >= 2 else None
                img_os = inspect_data[0].get("Os") or inspect_data[0].get("os")
                img_arch = inspect_data[0].get("Architecture") or inspect_data[0].get("architecture")
                # Normalize arch for comparison (e.g. x86_64 -> amd64, aarch64 -> arm64)
                arch_map = {"x86_64": "amd64", "aarch64": "arm64"}
                if img_arch and img_arch in arch_map:
                    img_arch = arch_map[img_arch]
                if required_os and img_os != required_os:
                    return None
                if required_arch and img_arch != required_arch:
                    return None

        return images

    def _should_rebuild_image(self, existing_image):
        """Determine if an image should be rebuilt based on Containerfile changes."""
        if not existing_image:
            return True

        if self.params.get("force"):
            return True

        # Check Containerfile hash
        containerfile_contents = self.containerfile_processor.get_containerfile_contents()
        if not containerfile_contents:
            return False

        current_hash = self.containerfile_processor.hash_containerfile_contents(containerfile_contents)
        if not current_hash:
            return False

        # Get existing hash from image labels
        existing_hash = ""
        if existing_image:
            labels = existing_image[0].get("Labels") or {}
            existing_hash = labels.get("containerfile.hash", "")

        return current_hash != existing_hash

    def present(self):
        """Ensure image is present (pull or build if needed)."""
        image = self.find_image()

        # Get digest before any operations
        digest_before = None
        if image:
            inspect_data = self.inspector.inspect_image(self.repository.full_name)
            if inspect_data and len(inspect_data) > 0:
                digest_before = inspect_data[0].get("Digest") or inspect_data[0].get("digest")
                # If Digest is not available, try to get from RepoDigests
                if not digest_before:
                    repo_digests = inspect_data[0].get("RepoDigests", [])
                    if repo_digests:
                        # Extract digest from first RepoDigest (format: repo@sha256:...)
                        for repo_digest in repo_digests:
                            if "@" in repo_digest:
                                digest_before = repo_digest.split("@", 1)[1]
                                break

        if not image or self._should_rebuild_image(image):
            if self.params.get("state") == "build" or self.params.get("path"):
                self._build_image()
            else:
                self._pull_image(digest_before)

        if self.params.get("push"):
            self._push_image()

        # Update results with final image info
        final_image = self.find_image()
        if final_image:
            self.results["image"] = self.inspector.inspect_image(self.repository.full_name)

    def _build_image(self):
        """Build an image."""
        build_config = self.params.get("build", {})
        path = self.params.get("path")

        # Validate build configuration
        if build_config.get("file") and build_config.get("container_file"):
            self.module.fail_json(msg="Cannot specify both build file and container file content!")

        if not path and not build_config.get("file") and not build_config.get("container_file"):
            self.module.fail_json(msg="Path to build context or file is required when building an image")

        # Get containerfile hash
        containerfile_contents = self.containerfile_processor.get_containerfile_contents()
        containerfile_hash = self.containerfile_processor.hash_containerfile_contents(containerfile_contents)

        # Build the image
        if not self.module.check_mode:
            image_id, output, podman_command = self.builder.build_image(
                self.repository.full_name, build_config, path, containerfile_hash,
                self.params.get("platform"),
                self.params.get("arch"),
            )
            self.results["stdout"] = output
            self.results["image"] = self.inspector.inspect_image(image_id)
            self.results["podman_actions"].append(podman_command)

        self.results["changed"] = True
        self.results["actions"].append(f"Built image {self.repository.full_name} from {path or 'context'}")

    def _pull_image(self, digest_before=None):
        """Pull an image."""
        if not self.params.get("pull", True):
            self.module.fail_json(msg=f"Image {self.repository.full_name} not found locally and pull is disabled")

        if not self.module.check_mode:
            unused, podman_command = self.puller.pull_image(
                self.repository.full_name,
                self.params.get("arch"),
                self.params.get("platform"),
                self.params.get("pull_extra_args"),
            )
            self.results["image"] = self.inspector.inspect_image(self.repository.full_name)
            self.results["podman_actions"].append(podman_command)

            # Check if digest actually changed
            digest_after = None
            if self.results["image"] and len(self.results["image"]) > 0:
                digest_after = self.results["image"][0].get("Digest") or self.results["image"][0].get("digest")
                # If Digest is not available, try to get from RepoDigests
                if not digest_after:
                    repo_digests = self.results["image"][0].get("RepoDigests", [])
                    if repo_digests:
                        # Extract digest from first RepoDigest (format: repo@sha256:...)
                        for repo_digest in repo_digests:
                            if "@" in repo_digest:
                                digest_after = repo_digest.split("@", 1)[1]
                                break

            changed = digest_before != digest_after
            self.results["changed"] = changed
            if changed:
                self.results["actions"].append(f"Pulled image {self.repository.full_name}")
        else:
            self.results["changed"] = True
            self.results["actions"].append(f"Pulled image {self.repository.full_name}")

    def _push_image(self):
        """Push an image."""
        push_config = self.params.get("push_args", {})

        if not self.module.check_mode:
            output, action = self.pusher.push_image(self.repository.full_name, push_config)
            self.results["stdout"] += "\n" + output
            self.results["actions"].append(action)
            self.results["podman_actions"].append(f"{self.executable} {action}")

        self.results["changed"] = True
        self.results["actions"].append(f"Pushed image {self.repository.full_name}")

    def absent(self):
        """Ensure image is absent."""
        image = self.find_image()

        if image:
            if not self.module.check_mode:
                unused, podman_command = self.remover.remove_image(
                    self.repository.full_name, self.params.get("force", False)
                )
                self.results["podman_actions"].append(podman_command)

            self.results["changed"] = True
            self.results["actions"].append(f"Removed image {self.repository.full_name}")
            self.results["image"] = {"state": "Deleted"}
        else:
            # Try removing by ID if provided
            image_id = self._extract_image_id(self.repository.original_name)
            if image_id and self._image_id_exists(image_id):
                if not self.module.check_mode:
                    unused, podman_command = self.remover.remove_image_by_id(image_id, self.params.get("force", False))
                    self.results["podman_actions"].append(podman_command)

                self.results["changed"] = True
                self.results["actions"].append(f"Removed image with ID {image_id}")
                self.results["image"] = {"state": "Deleted"}

    def _extract_image_id(self, name):
        """Extract image ID from name if it looks like an ID."""
        # Remove tag if present
        if ":" in name and not name.startswith("sha256:"):
            name = name.split(":")[0]
        return name

    def _image_id_exists(self, image_id):
        """Check if an image ID exists."""
        args = ["image", "ls", "--quiet", "--no-trunc"]
        rc, out, err = run_podman_command(self.module, self.executable, args, ignore_errors=True)

        if rc != 0:
            return False

        candidates = [line.replace("sha256:", "") for line in out.splitlines()]
        return any(candidate.startswith(image_id) for candidate in candidates)

    def execute(self):
        """Execute the requested operation."""
        state = self.params.get("state", "present")

        if state in ["present", "build"]:
            self.present()
        elif state == "absent":
            self.absent()
        elif state == "quadlet":
            # Quadlet functionality will be handled by the main module
            pass

        return self.results



#!/usr/bin/python

# Copyright (c) 2016, William L Thomson Jr
# Copyright (c) 2013, Yap Sok Ann
# Written by Yap Sok Ann <sokann@gmail.com>
# Modified by William L. Thomson Jr. <wlt@o-sinc.com>
# Based on apt module written by Matthew Williams <matthew@flowroute.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: portage
short_description: Package manager for Gentoo
description:
  - Manages Gentoo packages.
extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  package:
    description:
      - Package atom or set, for example V(sys-apps/foo) or V(>foo-2.13) or V(@world).
    aliases: [name]
    type: list
    elements: str

  state:
    description:
      - State of the package atom.
    default: "present"
    choices: ["present", "installed", "emerged", "absent", "removed", "unmerged", "latest"]
    type: str

  update:
    description:
      - Update packages to the best version available (C(--update)).
    type: bool
    default: false

  backtrack:
    description:
      - Set backtrack value (C(--backtrack)).
    type: int
    version_added: 5.8.0

  deep:
    description:
      - Consider the entire dependency tree of packages (C(--deep)).
    type: bool
    default: false

  newuse:
    description:
      - Include installed packages where USE flags have changed (C(--newuse)).
    type: bool
    default: false

  changed_deps:
    description:
      - Tells emerge to replace installed packages for which the ebuild dependencies
        have changed since the packages were built (C(--changed-deps)).
    type: bool
    default: false
    version_added: 12.0.0

  changed_use:
    description:
      - Include installed packages where USE flags have changed, except when.
      - Flags that the user has not enabled are added or removed.
      - (C(--changed-use)).
    type: bool
    default: false

  oneshot:
    description:
      - Do not add the packages to the world file (C(--oneshot)).
    type: bool
    default: false

  noreplace:
    description:
      - Do not re-emerge installed packages (C(--noreplace)).
    type: bool
    default: true

  nodeps:
    description:
      - Only merge packages but not their dependencies (C(--nodeps)).
    type: bool
    default: false

  onlydeps:
    description:
      - Only merge packages' dependencies but not the packages (C(--onlydeps)).
    type: bool
    default: false

  depclean:
    description:
      - Remove packages not needed by explicitly merged packages (C(--depclean)).
      - If no package is specified, clean up the world's dependencies.
      - Otherwise, C(--depclean) serves as a dependency aware version of C(--unmerge).
    type: bool
    default: false

  quiet:
    description:
      - Run emerge in quiet mode (C(--quiet)).
    type: bool
    default: false

  verbose:
    description:
      - Run emerge in verbose mode (C(--verbose)).
    type: bool
    default: false

  select:
    description:
      - If set to V(true), explicitely add the package to the world file.
      - Please note that this option is not used for idempotency, it is only used when actually installing a package.
    type: bool
    version_added: 8.6.0

  sync:
    description:
      - Sync package repositories first.
      - If V(yes), perform C(emerge --sync).
      - If V(web), perform C(emerge-webrsync).
    choices: ["web", "yes", "no"]
    type: str

  getbinpkgonly:
    description:
      - Merge only packages specified at C(PORTAGE_BINHOST) in C(make.conf).
    type: bool
    default: false
    version_added: 1.3.0

  getbinpkg:
    description:
      - Prefer packages specified at C(PORTAGE_BINHOST) in C(make.conf).
    type: bool
    default: false

  usepkgonly:
    description:
      - Merge only binaries (no compiling).
    type: bool
    default: false

  usepkg:
    description:
      - Tries to use the binary package(s) in the locally available packages directory.
    type: bool
    default: false

  keepgoing:
    description:
      - Continue as much as possible after an error.
    type: bool
    default: false

  jobs:
    description:
      - Specifies the number of packages to build simultaneously.
      - 'Since version 2.6: Value of V(0) or V(false) resets any previously added C(--jobs) setting values.'
    type: int

  loadavg:
    description:
      - Specifies that no new builds should be started if there are other builds running and the load average is at least
        LOAD.
      - 'Since version 2.6: Value of 0 or False resets any previously added C(--load-average) setting values.'
    type: float

  withbdeps:
    description:
      - Specifies that build time dependencies should be installed.
    type: bool
    version_added: 5.8.0

  quietbuild:
    description:
      - Redirect all build output to logs alone, and do not display it on stdout (C(--quiet-build)).
    type: bool
    default: false

  quietfail:
    description:
      - Suppresses display of the build log on stdout (--quiet-fail).
      - Only the die message and the path of the build log are displayed on stdout.
    type: bool
    default: false

author:
  - "William L Thomson Jr (@wltjr)"
  - "Yap Sok Ann (@sayap)"
  - "Andrew Udvare (@Tatsh)"
"""

EXAMPLES = r"""
- name: Make sure package foo is installed
  community.general.portage:
    package: foo
    state: present

- name: Make sure package foo is not installed
  community.general.portage:
    package: foo
    state: absent

- name: Update package foo to the latest version (os specific alternative to latest)
  community.general.portage:
    package: foo
    update: true

- name: Install package foo using PORTAGE_BINHOST setup
  community.general.portage:
    package: foo
    getbinpkg: true

- name: Re-install world from binary packages only and do not allow any compiling
  community.general.portage:
    package: '@world'
    usepkgonly: true

- name: Sync repositories and update world
  community.general.portage:
    package: '@world'
    update: true
    deep: true
    sync: true

- name: Remove unneeded packages
  community.general.portage:
    depclean: true

- name: Remove package foo if it is not explicitly needed
  community.general.portage:
    package: foo
    state: absent
    depclean: true
"""

import os
import re
import sys
import traceback

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils.common.respawn import has_respawned, respawn_module

try:
    from portage.dbapi import vartree
    from portage.exception import InvalidAtom

    HAS_PORTAGE = True
    PORTAGE_IMPORT_ERROR = None
except ImportError:
    HAS_PORTAGE = False
    PORTAGE_IMPORT_ERROR = traceback.format_exc()


def query_package(module, package, action):
    if package.startswith("@"):
        return query_set(module, package, action)
    return query_atom(module, package, action)


def query_atom(module, atom, action):
    vdb = vartree.vardbapi()
    try:
        exists = vdb.match(atom)
    except InvalidAtom:
        return False
    return bool(exists)


def query_set(module, set_, action):
    system_sets = [
        "@live-rebuild",
        "@module-rebuild",
        "@preserved-rebuild",
        "@security",
        "@selected",
        "@system",
        "@world",
        "@x11-module-rebuild",
    ]

    if set_ in system_sets:
        if action == "unmerge":
            module.fail_json(msg=f"set {set_} cannot be removed")
        return False

    world_sets_path = "/var/lib/portage/world_sets"
    if not os.path.exists(world_sets_path):
        return False

    cmd = ["grep", set_, world_sets_path]

    rc, out, err = module.run_command(cmd)
    return rc == 0


def sync_repositories(module, webrsync=False):
    if module.check_mode:
        module.exit_json(msg="check mode not supported by sync")

    if webrsync:
        webrsync_path = module.get_bin_path("emerge-webrsync", required=True)
        cmd = [webrsync_path, "--quiet"]
    else:
        cmd = [module.emerge_path, "--sync", "--quiet", "--ask=n"]

    rc, out, err = module.run_command(cmd)
    if rc != 0:
        module.fail_json(msg="could not sync package repositories")


# Note: In the 3 functions below, package querying is done one-by-one,
# but emerge is done in one go. If that is not desirable, split the
# packages into multiple tasks instead of joining them together with
# comma.


def emerge_packages(module, packages):
    """Run emerge command against given list of atoms."""
    p = module.params

    if (
        p["noreplace"]
        and not p["changed_deps"]
        and not p["changed_use"]
        and not p["newuse"]
        and not (p["update"] or p["state"] == "latest")
    ):
        for package in packages:
            if (
                p["noreplace"]
                and not p["changed_deps"]
                and not p["changed_use"]
                and not p["newuse"]
                and not query_package(module, package, "emerge")
            ):
                break
        else:
            module.exit_json(changed=False, msg="Packages already present.")
        if module.check_mode:
            module.exit_json(changed=True, msg="Packages would be installed.")

    args = []
    emerge_flags = {
        "update": "--update",
        "deep": "--deep",
        "newuse": "--newuse",
        "changed_deps": "--changed-deps",
        "changed_use": "--changed-use",
        "oneshot": "--oneshot",
        "noreplace": "--noreplace",
        "nodeps": "--nodeps",
        "onlydeps": "--onlydeps",
        "quiet": "--quiet",
        "verbose": "--verbose",
        "getbinpkgonly": "--getbinpkgonly",
        "getbinpkg": "--getbinpkg",
        "usepkgonly": "--usepkgonly",
        "usepkg": "--usepkg",
        "keepgoing": "--keep-going",
        "quietbuild": "--quiet-build",
        "quietfail": "--quiet-fail",
    }
    for flag, arg in emerge_flags.items():
        if p[flag]:
            args.append(arg)

    if p["state"] and p["state"] == "latest":
        args.append("--update")

    emerge_flags = {
        "jobs": "--jobs",
        "loadavg": "--load-average",
        "backtrack": "--backtrack",
        "withbdeps": "--with-bdeps",
        "select": "--select",
    }

    for flag, arg in emerge_flags.items():
        flag_val = p[flag]

        if flag_val is None:
            """Fallback to default: don't use this argument at all."""
            continue

        """Add the --flag=value pair."""
        if isinstance(flag_val, bool):
            args.extend((arg, "y" if flag_val else "n"))
        elif not flag_val:
            """If the value is 0 or 0.0: add the flag, but not the value."""
            args.append(arg)
        else:
            args.extend((arg, str(flag_val)))

    cmd, (rc, out, err) = run_emerge(module, packages, *args)
    if rc != 0:
        module.fail_json(
            cmd=cmd,
            rc=rc,
            stdout=out,
            stderr=err,
            msg="Packages not installed.",
        )

    # Check for SSH error with PORTAGE_BINHOST, since rc is still 0 despite
    #   this error
    if (p["usepkgonly"] or p["getbinpkg"] or p["getbinpkgonly"]) and "Permission denied (publickey)." in err:
        module.fail_json(
            cmd=cmd,
            rc=rc,
            stdout=out,
            stderr=err,
            msg="Please check your PORTAGE_BINHOST configuration in make.conf and your SSH authorized_keys file",
        )

    changed = True
    for line in out.splitlines():
        if re.match(r"(?:>+) Emerging (?:binary )?\(1 of", line):
            msg = "Packages installed."
            break
        elif module.check_mode and re.match(r"\[(binary|ebuild)", line):
            msg = "Packages would be installed."
            break
    else:
        changed = False
        msg = "No packages installed."

    module.exit_json(
        changed=changed,
        cmd=cmd,
        rc=rc,
        stdout=out,
        stderr=err,
        msg=msg,
    )


def unmerge_packages(module, packages):
    p = module.params

    for package in packages:
        if query_package(module, package, "unmerge"):
            break
    else:
        module.exit_json(changed=False, msg="Packages already absent.")

    args = ["--unmerge"]

    for flag in ["quiet", "verbose"]:
        if p[flag]:
            args.append(f"--{flag}")

    cmd, (rc, out, err) = run_emerge(module, packages, *args)

    if rc != 0:
        module.fail_json(
            cmd=cmd,
            rc=rc,
            stdout=out,
            stderr=err,
            msg="Packages not removed.",
        )

    module.exit_json(
        changed=True,
        cmd=cmd,
        rc=rc,
        stdout=out,
        stderr=err,
        msg="Packages removed.",
    )


def cleanup_packages(module, packages):
    p = module.params

    if packages:
        for package in packages:
            if query_package(module, package, "unmerge"):
                break
        else:
            module.exit_json(changed=False, msg="Packages already absent.")

    args = ["--depclean"]

    for flag in ["quiet", "verbose"]:
        if p[flag]:
            args.append(f"--{flag}")

    cmd, (rc, out, err) = run_emerge(module, packages, *args)
    if rc != 0:
        module.fail_json(
            cmd=cmd,
            rc=rc,
            stdout=out,
            stderr=err,
            msg="Packages not cleaned up.",
        )

    removed = 0
    for line in out.splitlines():
        if not line.startswith("Number removed:"):
            continue
        parts = line.split(":")
        removed = int(parts[1].strip())
    changed = removed > 0

    module.exit_json(
        changed=changed,
        cmd=cmd,
        rc=rc,
        stdout=out,
        stderr=err,
        msg="Depclean completed.",
    )


def run_emerge(module, packages, *args):
    args = list(args)

    args.append("--ask=n")
    if module.check_mode:
        args.append("--pretend")

    cmd = [module.emerge_path] + args + packages
    return cmd, module.run_command(cmd)


portage_present_states = ["present", "emerged", "installed", "latest"]
portage_absent_states = ["absent", "unmerged", "removed"]


def main():
    module = AnsibleModule(
        argument_spec=dict(
            package=dict(type="list", elements="str", aliases=["name"]),
            state=dict(
                default=portage_present_states[0],
                choices=portage_present_states + portage_absent_states,
            ),
            update=dict(default=False, type="bool"),
            backtrack=dict(type="int"),
            deep=dict(default=False, type="bool"),
            newuse=dict(default=False, type="bool"),
            changed_deps=dict(default=False, type="bool"),
            changed_use=dict(default=False, type="bool"),
            oneshot=dict(default=False, type="bool"),
            noreplace=dict(default=True, type="bool"),
            nodeps=dict(default=False, type="bool"),
            onlydeps=dict(default=False, type="bool"),
            depclean=dict(default=False, type="bool"),
            select=dict(type="bool"),
            quiet=dict(default=False, type="bool"),
            verbose=dict(default=False, type="bool"),
            sync=dict(choices=["yes", "web", "no"]),
            getbinpkgonly=dict(default=False, type="bool"),
            getbinpkg=dict(default=False, type="bool"),
            usepkgonly=dict(default=False, type="bool"),
            usepkg=dict(default=False, type="bool"),
            keepgoing=dict(default=False, type="bool"),
            jobs=dict(type="int"),
            loadavg=dict(type="float"),
            withbdeps=dict(type="bool"),
            quietbuild=dict(default=False, type="bool"),
            quietfail=dict(default=False, type="bool"),
        ),
        required_one_of=[["package", "sync", "depclean"]],
        mutually_exclusive=[
            ["nodeps", "onlydeps"],
            ["quiet", "verbose"],
            ["quietbuild", "verbose"],
            ["quietfail", "verbose"],
            ["oneshot", "select"],
        ],
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    if not HAS_PORTAGE:
        if sys.executable != "/usr/bin/python" and not has_respawned():
            respawn_module("/usr/bin/python")
        else:
            module.fail_json(msg=missing_required_lib("portage"), exception=PORTAGE_IMPORT_ERROR)

    module.emerge_path = module.get_bin_path("emerge", required=True)

    p = module.params

    if p["sync"] and p["sync"].strip() != "no":
        sync_repositories(module, webrsync=(p["sync"] == "web"))
        if not p["package"]:
            module.exit_json(msg="Sync successfully finished.")

    packages = []
    if p["package"]:
        packages.extend(p["package"])

    if p["depclean"]:
        if packages and p["state"] not in portage_absent_states:
            module.fail_json(
                msg=f"Depclean can only be used with package when the state is one of: {portage_absent_states}",
            )

        cleanup_packages(module, packages)

    elif p["state"] in portage_present_states:
        emerge_packages(module, packages)

    elif p["state"] in portage_absent_states:
        unmerge_packages(module, packages)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_topology_link_layer_info
short_description: Information module for networks _topology _link _layer
description:
  - Get all networks _topology _link _layer. - > List the LLDP and CDP information
    for all discovered devices and connections in a network. At least one MX or MS
    device must be in the network in order to build the topology.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  networkId:
    description:
      - NetworkId path parameter. Network ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for networks getNetworkTopologyLinkLayer
    description: Complete reference of the getNetworkTopologyLinkLayer API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-network-topology-link-layer
notes:
  - SDK Method used are
    networks.Networks.get_network_topology_link_layer,
  - Paths used are
    get /networks/{networkId}/topology/linkLayer,
"""

EXAMPLES = r"""
- name: Get all networks _topology _link _layer
  cisco.meraki.networks_topology_link_layer_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "nodes": [
        {
          "derivedId": "string",
          "mac": "string",
          "type": "string",
          "root": true,
          "device": {
            "serial": "string",
            "name": "string",
            "model": "string",
            "productType": "string",
            "status": "string",
            "lastReportedAt": "string",
            "clients": {
              "counts": {
                "total": 0
              }
            },
            "uplinks": [
              {
                "vlanId": 0
              }
            ]
          },
          "discovered": {
            "lldp": {
              "chassisId": "string",
              "systemName": "string",
              "systemDescription": "string",
              "systemCapabilities": [
                "string"
              ],
              "managementAddress": "string"
            },
            "cdp": {
              "platform": "string",
              "deviceId": "string",
              "address": "string",
              "capabilities": [
                "string"
              ],
              "managementAddress": "string"
            }
          },
          "stack": {
            "id": 0,
            "name": "string",
            "members": [
              {}
            ],
            "clients": {
              "counts": {
                "total": 0
              }
            }
          }
        }
      ],
      "links": [
        {
          "ends": [
            {
              "node": {
                "derivedId": "string",
                "type": "string"
              },
              "device": {
                "serial": "string",
                "name": "string"
              },
              "discovered": {
                "lldp": {
                  "portId": "string",
                  "portDescription": "string"
                },
                "cdp": {
                  "portId": "string",
                  "nativeVlan": 0
                }
              }
            }
          ],
          "lastReportedAt": "string"
        }
      ],
      "errors": [
        "string"
      ]
    }
"""



#!/usr/bin/python

# Copyright (c) 2013, Patrick Pelletier <pp.pelletier@gmail.com>
# Based on pacman (Afterburn) and pkgin (Shaun Zinck) modules
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: opkg
author: "Patrick Pelletier (@skinp)"
short_description: Package manager for OpenWrt and Openembedded/Yocto based Linux distributions
description:
  - Manages ipk packages for OpenWrt and Openembedded/Yocto based Linux distributions.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of package(s) to install/remove.
      - V(NAME=VERSION) syntax is also supported to install a package in a certain version. See the examples. This only works
        on Yocto based Linux distributions (opkg>=0.3.2) and not for OpenWrt. This is supported since community.general 6.2.0.
    aliases: [pkg]
    required: true
    type: list
    elements: str
  state:
    description:
      - State of the package.
    choices: ['present', 'absent', 'installed', 'removed']
    default: present
    type: str
  force:
    description:
      - The C(opkg --force) parameter used.
    choices:
      - "depends"
      - "maintainer"
      - "reinstall"
      - "overwrite"
      - "downgrade"
      - "space"
      - "postinstall"
      - "remove"
      - "checksum"
      - "removal-of-dependent-packages"
    type: str
  update_cache:
    description:
      - Update the package DB first.
    default: false
    type: bool
  executable:
    description:
      - The executable location for C(opkg).
    type: path
    version_added: 7.2.0
requirements:
  - opkg
  - python
"""

EXAMPLES = r"""
- name: Install foo
  community.general.opkg:
    name: foo
    state: present

- name: Install foo in version 1.2 (opkg>=0.3.2 on Yocto based Linux distributions)
  community.general.opkg:
    name: foo=1.2
    state: present

- name: Update cache and install foo
  community.general.opkg:
    name: foo
    state: present
    update_cache: true

- name: Remove foo
  community.general.opkg:
    name: foo
    state: absent

- name: Remove foo and bar
  community.general.opkg:
    name:
      - foo
      - bar
    state: absent

- name: Install foo using overwrite option forcibly
  community.general.opkg:
    name: foo
    state: present
    force: overwrite
"""

RETURN = r"""
version:
  description: Version of opkg.
  type: str
  returned: always
  sample: "2.80.0"
  version_added: 10.0.0
"""

import os

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt
from ansible_collections.community.general.plugins.module_utils._module_helper import StateModuleHelper


class Opkg(StateModuleHelper):
    module = dict(
        argument_spec=dict(
            name=dict(aliases=["pkg"], required=True, type="list", elements="str"),
            state=dict(default="present", choices=["present", "installed", "absent", "removed"]),
            force=dict(
                choices=[
                    "depends",
                    "maintainer",
                    "reinstall",
                    "overwrite",
                    "downgrade",
                    "space",
                    "postinstall",
                    "remove",
                    "checksum",
                    "removal-of-dependent-packages",
                ]
            ),
            update_cache=dict(default=False, type="bool"),
            executable=dict(type="path"),
        ),
    )

    def __init_module__(self):
        self.vars.set("install_c", 0, output=False, change=True)
        self.vars.set("remove_c", 0, output=False, change=True)

        state_map = dict(
            query="list-installed",
            present="install",
            installed="install",
            absent="remove",
            removed="remove",
        )

        dir, cmd = os.path.split(self.vars.executable) if self.vars.executable else (None, "opkg")

        self.runner = CmdRunner(
            self.module,
            command=cmd,
            arg_formats=dict(
                package=cmd_runner_fmt.as_list(),
                state=cmd_runner_fmt.as_map(state_map),
                force=cmd_runner_fmt.as_optval("--force-"),
                update_cache=cmd_runner_fmt.as_bool("update"),
                version=cmd_runner_fmt.as_fixed("--version"),
            ),
            path_prefix=[dir],
        )

        with self.runner("version") as ctx:
            rc, out, err = ctx.run()
            self.vars.version = out.strip().replace("opkg version ", "")

        if self.vars.update_cache:
            rc, dummy, dummy = self.runner("update_cache").run()
            if rc != 0:
                self.do_raise("could not update package db")

    @staticmethod
    def split_name_and_version(package):
        """Split the name and the version when using the NAME=VERSION syntax"""
        splitted = package.split("=", 1)
        if len(splitted) == 1:
            return splitted[0], None
        else:
            return splitted[0], splitted[1]

    def _package_in_desired_state(self, name, want_installed, version=None):
        dummy, out, dummy = self.runner("state package").run(state="query", package=name)

        has_package = out.startswith(f"{name} - {'' if not version else f'{version} '}")
        return want_installed == has_package

    def state_present(self):
        with self.runner("state force package") as ctx:
            for package in self.vars.name:
                pkg_name, pkg_version = self.split_name_and_version(package)
                if (
                    not self._package_in_desired_state(pkg_name, want_installed=True, version=pkg_version)
                    or self.vars.force == "reinstall"
                ):
                    ctx.run(package=package)
                    self.vars.set("run_info", ctx.run_info, verbosity=4)
                    if not self._package_in_desired_state(pkg_name, want_installed=True, version=pkg_version):
                        self.do_raise(f"failed to install {package}")
                    self.vars.install_c += 1
        if self.vars.install_c > 0:
            self.vars.msg = f"installed {self.vars.install_c} package(s)"
        else:
            self.vars.msg = "package(s) already present"

    def state_absent(self):
        with self.runner("state force package") as ctx:
            for package in self.vars.name:
                package, dummy = self.split_name_and_version(package)
                if not self._package_in_desired_state(package, want_installed=False):
                    ctx.run(package=package)
                    self.vars.set("run_info", ctx.run_info, verbosity=4)
                    if not self._package_in_desired_state(package, want_installed=False):
                        self.do_raise(f"failed to remove {package}")
                    self.vars.remove_c += 1
        if self.vars.remove_c > 0:
            self.vars.msg = f"removed {self.vars.remove_c} package(s)"
        else:
            self.vars.msg = "package(s) already absent"

    state_installed = state_present
    state_removed = state_absent


def main():
    Opkg.execute()


if __name__ == "__main__":
    main()
